/*
 * module.c - module interface for a block device probe
 *
 * Copyright 2005-2008 Red Hat, Inc.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author(s): Peter Jones <pjones@redhat.com>
 */

#define _GNU_SOURCE 1

#include <sys/types.h>
#include <dirent.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dlfcn.h>
#include <errno.h>
#include <argz.h>

#include "bdevid.h"

/* probe registration */

int bdevid_register_probe(struct bdevid_module *m, struct bdevid_probe_ops *ops)
{
    struct bdevid_probe *probe, **p;
    
    if (!(probe = calloc(1, sizeof (*probe))))
        return -1;

    p = (struct bdevid_probe **)realloc(m->probes, (m->probes_count+1)*sizeof(struct bdevid_probe *));
    if (!p) {
       return -1;
    }
    m->probes = p;
    m->probes[m->probes_count] = probe;
    m->probes_count++;
    probe->module = m;
    switch (m->version) {
        case 1:
            probe->ops.ops_v1 = (struct bdevid_probe_ops_v1 *)ops;
            probe->name = probe->ops.ops_v1->name;
            break;
        case 2:
            probe->ops.ops = ops;
            probe->name = probe->ops.ops->name;
            break;
    }

    return 0;
}

/* end probe registration */
/* begin loading and unloading */

static struct bdevid_module * bdevid_module_lookup(struct bdevid *b, char *name)
{
    int i;
    if(!b->modules)
        return NULL;
    for(i=0; i<b->modules_count; i++)
        if(!strcmp(b->modules[i]->name, name))
            return b->modules[i];
    return NULL;
}

static void bdevid_module_unlist(struct bdevid *b, char *name)
{
    struct bdevid_module **p;
    int i;
    if(!b->modules)
        return;
    for(i=0; i<b->modules_count; i++)
        if(!strcmp(b->modules[i]->name, name)) {
            b->modules_count--;
            if(i<b->modules_count) {
                 memmove(b->modules[i], b->modules[i+1], b->modules_count-i);
             }
             p = (struct bdevid_module **)realloc(b->modules, b->modules_count*sizeof(struct bdevid_module *));
             if(p || !b->modules_count)
                 b->modules = p;
             return;
        }
}

int bdevid_module_unload(struct bdevid *b, char *name)
{
    struct bdevid_module *mod = NULL, **p;
    int i;

    mod = bdevid_module_lookup(b, name);
    if (mod) {
        if (mod->probes) {
            for(i=0; i<mod->probes_count; i++)
                free(mod->probes[i]);
            free(mod->probes);
        }

        if (mod->bdevid)
            for(i=0; i<b->modules_count; i++)
                if(!strcmp(b->modules[i]->name, mod->name)) {
                    b->modules_count--;
                    if(i<b->modules_count) {
                        memmove(b->modules[i], b->modules[i+1], b->modules_count-i);
                        i--;
                    }
                    p = (struct bdevid_module **)realloc(b->modules, b->modules_count*sizeof(struct bdevid_module *));
                    if(p || !b->modules_count)
                        b->modules = p;
                }
        if (mod->dlh)
            dlclose(mod->dlh);

        free(mod);
    }

    return 0;
}

int bdevid_module_load_file_maybe(struct bdevid *b, char *file, char *name)
{
    struct bdevid_module *m, **p;
    struct bdevid_module_ops **ops;
    uint32_t *version;
    int rc = -1;
    int i;

    if (access(file, X_OK))
        return -1;

    if (!(m = calloc(1, sizeof (*m))))
        return -1;

    dlerror();
    if (!(m->dlh = dlopen(file, RTLD_NOW|RTLD_LOCAL)))
        goto out;

    dlerror();
    if (!(version = dlsym(m->dlh, "bdevid_module_version")))
        m->version = 1;
    else if (*version >= 0x2 && *version <= BDEVID_MODULE_VERSION)
        m->version = *version;
    else
        goto out;

    dlerror();
    if (!(ops = dlsym(m->dlh, "bdevid_module_ops")) || !*ops)
        goto out;

    m->ops = *ops;
    if (m->ops->magic != BDEVID_MAGIC)
        goto out;

    m->name = m->ops->name;
    if (name && strcmp(m->name, name))
        goto out;

    for(i=0; i<b->modules_count; i++)
        if(!strcmp(b->modules[i]->name, m->name)) {
            rc = 0;
            goto out;
        }

    p = (struct bdevid_module **)realloc(b->modules, (b->modules_count+1)*sizeof(struct bdevid_module *));
    if (!p)
        goto out;
    b->modules = p;

    b->modules[b->modules_count] = m;
    b->modules_count++;
    m->bdevid = b;
    m->probes = NULL;
    m->probes_count = 0;

    if (m->ops->init(m, bdevid_register_probe) < 0)
        goto out;

    return 0;
out:
    if (m) {
        int en = errno;

        dlerror();
        dlerror();

        if (m->bdevid)
            bdevid_module_unlist(b, m->name);

        if (m->dlh)
            dlclose(m->dlh);

        memset(m, '\0', sizeof (*m));
        free(m);
        errno = en;
    }
    m = NULL;
    if (!rc)
        errno = 0;
    return rc;
}

int bdevid_module_load_file(struct bdevid *b, char *file)
{
    return bdevid_module_load_file_maybe(b, file, NULL);
    
}

int bdevid_module_load(struct bdevid *b, char *name)
{
    char *dir;

    if (bdevid_module_lookup(b, name))
        return 0;

    for (dir = b->module_pathz; dir;
            dir = argz_next(b->module_pathz, b->module_pathz_len, dir)) {
        char *file = NULL;

        if (asprintf(&file, "%s/%s.so", dir, name) < 0)
            continue;

        if (bdevid_module_load_file_maybe(b, file, name) < 0) {
            free(file);
            continue;
        }
        free(file);
        return 0;
    }

    return -1;
}

int bdevid_module_load_all(struct bdevid *b)
{
    char *cur;

    for (cur = b->module_pathz; cur; 
            cur = argz_next(b->module_pathz, b->module_pathz_len, cur)) {
        DIR *dir = NULL;
        struct dirent *dent = NULL;

        if (!(dir = opendir(cur)))
            continue;

        while ((dent = readdir(dir))) {
            char *file = NULL;
            if (!strcmp(dent->d_name, ".") || !strcmp(dent->d_name, ".."))
                continue;

            if (asprintf(&file, "%s/%s", cur, dent->d_name) < 0)
                continue;

            bdevid_module_load_file(b, file);

            free(file);
        }
        closedir(dir);
    }

    return 0;
}

int bdevid_module_unload_all(struct bdevid *b)
{
    int i, j;
    for(i=0; i<b->modules_count; i++) {
        if (b->modules[i]->probes) {
            for(j=0; j<b->modules[i]->probes_count; j++)
                free(b->modules[i]->probes[j]);
            free(b->modules[i]->probes);
        }

        if (b->modules[i]->dlh)
            dlclose(b->modules[i]->dlh);

        free(b->modules[i]);
    }
    free(b->modules);
    b->modules = NULL;
    b->modules_count = 0;
    return 0;
}

/* end loading and unloading */

/*
 * vim:ts=8:sw=4:sts=4:et
 */
