/*
 * dm.c - backend library for partition table scanning on dm devices
 *
 * Copyright 2005-2008 Red Hat, Inc.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author(s): Peter Jones <pjones@redhat.com>
 */

#define _GNU_SOURCE 1
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <search.h>

#include <libdevmapper.h>
#include <parted/parted.h>

#include <nash.h>

#include "dm.h"
#include "lib.h"
#include "block.h"
#include "util.h"

static int
nashDmTaskNew(int type, const char *name, struct dm_task **task)
{
    int ret;

    *task = dm_task_create(type);
    if (!*task)
        return -1;
    if (name)
        dm_task_set_name(*task, name);
    ret = dm_task_run(*task);
    if (ret < 0) {
        dm_task_destroy(*task);
        *task = NULL;
    }
    return ret;
}

static int
nashDmGetInfo(const char *name, struct dm_task **task, struct dm_info *info)
{
    int ret;

    ret = nashDmTaskNew(DM_DEVICE_INFO, name, task);
    if (ret < 0)
        return ret;

    ret = dm_task_get_info(*task, info);
    if (ret < 0 || !info->exists) {
        dm_task_destroy(*task);
        *task = NULL;
        return 1;
    }
    return 0;
}

/* XXX this should be refactored so that it's given a specific map,
 * not just a name */
static char *
nashDmGetType(const char *name)
{
    void *next = NULL;
    char *type = NULL;
    int err;
    int ret;

    struct dm_task *task = NULL;

    ret = nashDmTaskNew(DM_DEVICE_TABLE, name, &task);
    if (ret < 0)
        return NULL;

    err = 0;
    ret = -1;
    do {
        uint64_t start, length;
        char *params;
        char *tmp = NULL;

        next = dm_get_next_target(task, next, &start, &length, &tmp, &params);
        if (tmp && !strcmp(tmp, "error")) {
            err = 1;
        } else if (!type && tmp) {
            type = strdup(tmp);
        }
    } while (next);

    dm_task_destroy(task);
    if (!type && err)
        return strdup("error");
    return type;
}

char *
nashDmGetUUID(const char *name)
{
    char *uuid = NULL;
    struct dm_task *task;
    struct dm_info info;

    if (!nashDmGetInfo(name, &task, &info)) {
        uuid = (char *)dm_task_get_uuid(task);
        if (uuid) {
            if (uuid[0] == '\0')
                uuid = NULL;
            else
                uuid = strdup(uuid);
        }
        dm_task_destroy(task);
    }

    return uuid;
}

dev_t
nashDmGetDev(const char *name)
{
    struct dm_task *task;
    struct dm_info info;
    dev_t ret = 0;

    task = dm_task_create(DM_DEVICE_INFO);
    if (!task)
        return 0;

    dm_task_set_name(task, name);
    dm_task_run(task);

    dm_task_get_info(task, &info);
    if (!info.exists) {
        if (task)
            dm_task_destroy(task);
        return 0;
    }

    ret = makedev(info.major, info.minor);
    dm_task_destroy(task);

    return ret;
}

static int
nashDmMapExists(const char *name)
{
    struct dm_task *task;
    struct dm_info info;

    if (nashDmGetInfo(name, &task, &info))
        return 0;
    dm_task_destroy(task);
    return 1;
}

int
nashDmCreate(nashContext *nc, char *name, char *uuid, long long start,
        long long length, char *type, char *params)
{
    char *real_params;
    char *dmname;
    struct dm_task *task;
    int rc;

    if (name && name[0] == '@')
        dmname = nash_dev_node_get_dm_name(nc, name);
    else
        dmname = name;

    if (!dmname) {
        //eprintf("Could not determine device name for %s\n", name);
        return 0;
    }

    if (nashDmMapExists(dmname))
        return 1;

    task = dm_task_create(DM_DEVICE_CREATE);
    if (!task)
        return 0;

    dm_task_set_name(task, dmname);
    if (uuid)
        dm_task_set_uuid(task, uuid);

#if 0
    printf("old rule: %s %Ld %Ld %s %s\n", dmname, start, length, type, params);
    real_params = nash_dev_tree_replace_dm_params(nc, start, length, params);
    printf("new rule: %s %Ld %Ld %s %s\n", dmname, start, length, type, real_params);
#else
    real_params = nash_dev_tree_replace_dm_params(nc, start, length, params);
#endif
    dm_task_add_target(task, start, length, type, real_params);
    free(real_params);

    rc = dm_task_run(task);
    dm_task_destroy(task);

    dm_task_update_nodes();

    if (rc < 0)
        return 0;

    return 1;
}

#if 0
int nashDmCloneAsError(char *oldname, char *newname)
{
    struct dm_task *task = NULL;
    int errnum;
    int ret;
    void *next = NULL;

    if ((ret = nashDmTaskNew(DM_DEVICE_TABLE, oldname, &task)) < 0)
        return -1;

    do {
        uint64_t start, length;
        char *params = NULL;
        char *type = NULL;

        next = dm_get_next_target(task, next, &start, &length, &type, &params);
        if (next) {
            ret = nashDmCreate(newname, NULL, start, length, "error", "");
            if (!ret)
                break;
        }
    } while (next);
    dm_task_destroy(task);
    return ret;
}
#endif

int
nashDmRemove(char *name)
{
    struct dm_task *task;
    int rc;

    task = dm_task_create(DM_DEVICE_REMOVE);
    if (!task)
        return 0;

    dm_task_set_name(task, name);

    rc = dm_task_run(task);
    dm_task_destroy(task);

    dm_task_update_nodes();

    if (rc < 0)
        return 0;

    return 1;
}

#if 0
int nashDmCloneAsError(char *oldname, char *newname)
{
    struct dm_task *task = NULL;
    int errnum;
    int ret;
    void *next = NULL;

    if ((ret = nashDmTaskNew(DM_DEVICE_TABLE, oldname, &task)) < 0)
        return -1;

    do {
        uint64_t start, length;
        char *params = NULL;
        char *type = NULL;

        next = dm_get_next_target(task, next, &start, &length, &type, &params);
        if (next) {
            ret = nashDmCreate(newname, NULL, start, length, "error", "");
            if (!ret)
                break;
        }
    } while (next);
    dm_task_destroy(task);
    return ret;
}
#endif

static int nashPartedError = 0;
static int nashPartedErrorDisplay = 1;

static nashContext *nash_parted_context = NULL;

static PedExceptionOption
nashPartedExceptionHandler(PedException *ex)
{
    nash_log_level level;

    switch (ex->type) {
        case PED_EXCEPTION_INFORMATION:
            nashPartedError = 0;
            level = NASH_NOTICE;
            break;
        case PED_EXCEPTION_WARNING:
            nashPartedError = 0;
            level = NASH_WARNING;
            break;
        default:
            level = NASH_ERROR;
            nashPartedError = 1;
            break;
    }
    if (nashPartedErrorDisplay)
        nashLogger(nash_parted_context, level, "%s\n", ex->message);
    switch (ex->options) {
        case PED_EXCEPTION_OK:
        case PED_EXCEPTION_CANCEL:
        case PED_EXCEPTION_IGNORE:
            return ex->options;
        default:
            break;
    }
    return PED_EXCEPTION_UNHANDLED;
}

int
nashDmCreatePartitions(nashContext *nc, char *path)
{
    PedDevice *dev;
    PedDisk *disk;
    PedPartition *part = NULL;
    PedExceptionHandler *old_handler;
    char *namestart;
    int nparts = 0;
    struct stat sb;
    char *parent_uuid;
    char *newpath = NULL;

    if (path[0] != '/') {
        if (asprintf(&newpath, "/dev/mapper/%s", path) == -1)
            return nparts;
        path = newpath;
    }
    if (stat(path, &sb) < 0 || !S_ISBLK(sb.st_mode))
        return nparts;

    namestart = strrchr(path, '/');
    if (!namestart || !*(namestart++))
        goto out;
    if (!*namestart)
        goto out;

    parent_uuid = nashDmGetUUID(namestart);

    old_handler = ped_exception_get_handler();
    nash_parted_context = nc;
    ped_exception_set_handler(nashPartedExceptionHandler);

    dev = ped_device_get(path);
    if (!dev || nashPartedError)
        goto out;

    if (!ped_device_open(dev))
        goto out;

    disk = ped_disk_new(dev);
    if (!disk || nashPartedError)
        goto out;

    part = ped_disk_next_partition(disk, NULL);
    while (part) {
        if (ped_partition_is_active(part))  {
            int i;
            char *name = NULL;
            char *table = NULL;
            char *uuid = NULL;

            i = asprintf(&name, "%sp%d", namestart, part->num);
            if (i < 0)
                continue;

            i = asprintf(&table, "%d:%d %Ld", major(sb.st_rdev),
                        minor(sb.st_rdev), part->geom.start);
            if (i < 0) {
                free(name);
                continue;
            }

            if (parent_uuid) {
                i = asprintf(&uuid, "%s-partition%d", parent_uuid, part->num);
                if (i < 0) {
                    free(name);
                    free(table);
                    continue;
                }
            }

            nparts += nashDmCreate(nc, name, uuid, 0, part->geom.length,
                    "linear", table);

            free(uuid);
            free(table);
            free(name);
        }
        part = ped_disk_next_partition(disk, part);
    }

out:
    dm_task_update_nodes();

    if (disk) {
        ped_disk_destroy(disk);
        ped_device_close(dev);
    }

    if (dev)
        ped_device_destroy(dev);

    if (parent_uuid)
        free(parent_uuid);
    
    if (newpath)
        free(newpath);

    ped_exception_set_handler(old_handler);
    nashPartedError = 0;
    return nparts;
}

/* ok, this whole data structure pretty much just sucks ass. */
struct dm_iter_object {
    const char *name;
    dev_t devno;
    char *type;

    int position;
    int visited;

    struct dm_iter_object **deps;
    int ndeps;
};

static int
dm_iter_object_namesort(const void *ov0, const void *ov1)
{
    const struct dm_iter_object *o0 = *(void **)ov0, *o1 = *(void **)ov1;

    if (!o0 || !o1)
        return (o0 ? 1 : 0) - (o1 ? 1 : 0);
    return strcmp(o0->name, o1->name);
}

static int
dm_iter_object_devsort(const void *ov0, const void *ov1)
{
    const struct dm_iter_object *o0 = *(void **)ov0, *o1 = *(void **)ov1;
    return o1->devno - o0->devno;
}

struct dm_iter {
    union {
        struct dm_iter_object **objects;
        char **names;
    };
    const char **prune_names;
    size_t nobjs;
    int i;
};

static struct dm_iter *
dm_iter_begin(const char **names)
{
    struct dm_iter *iter;
    struct dm_names *dmnames;
    int i = 0, j = 0;
    unsigned int next = 0;
    struct dm_task *task;

    iter = calloc(1, sizeof (struct dm_iter));
    iter->names = calloc(i+1, sizeof (void *));

    if (names && *names)
        iter->prune_names = names;

    task = dm_task_create(DM_DEVICE_LIST);
    if (!task) {
out:
        free(iter->names);
        free(iter);
        return NULL;
    }
    dm_task_run(task);
    if (!(dmnames = dm_task_get_names(task))) {
        dm_task_destroy(task);
        goto out;
    }
    do {
        dmnames = (void *)dmnames + next;

        if (dmnames->name) {
            iter->names = realloc(iter->names, sizeof (void *) * (i+1));
            iter->names[i++] = strdup(dmnames->name);
        }

        next = dmnames->next;
    } while (next);
    iter->nobjs = i;

    dm_task_destroy(task);
    for (i = 0; i < iter->nobjs; i++) {
        struct dm_iter_object *obj = calloc(1, sizeof (struct dm_iter_object));

        obj->name = iter->names[i];
        obj->devno = nashDmGetDev(obj->name);
        obj->type = nashDmGetType(obj->name);
        obj->position = i;

        iter->objects[i] = obj;
    }
    qsort(iter->objects, i, sizeof (struct dm_iter_object *),
            dm_iter_object_devsort);

    for (i = 0; i < iter->nobjs; i++) {
        struct dm_iter_object *obj = iter->objects[i];
        struct dm_deps *deps;
        struct dm_task *task;
        struct dm_info info;

        memset(&info, '\0', sizeof (info));

        if (nashDmGetInfo(obj->name, &task, &info) <= 0)
            continue;

        dm_task_destroy(task);

        if (nashDmTaskNew(DM_DEVICE_DEPS, obj->name, &task) < 0)
            continue;

        deps = dm_task_get_deps(task);

        obj->deps = calloc(1, sizeof (struct dm_iter_object *));
        obj->deps[0] = NULL;
        obj->ndeps = deps->count;
        for (j = 0; j < obj->ndeps; j++) {
            struct dm_iter_object **depp;
            struct dm_iter_object dummy = {
                .devno = deps->device[j]
            };
            struct dm_iter_object *dummyp = &dummy;

            obj->deps = realloc(obj->deps,
                sizeof (struct dm_iter_object *) * (j+1));
            depp = bsearch(&dummyp, iter->objects, iter->nobjs,
                sizeof (struct dm_iter_object *), dm_iter_object_devsort);
            if (!depp) {
                obj->ndeps--;
                j--;
                continue;
            }
            obj->deps[j] = *depp;
        }
        dm_task_destroy(task);
        qsort(obj->deps, obj->ndeps, sizeof (struct dm_iter_object *),
                dm_iter_object_devsort);
    }

    return iter;
}

static void
dm_iter_reset(struct dm_iter *iter)
{
    int i;

    for (i = 0; i < iter->nobjs; i++) {
        iter->objects[i]->visited = 0;
    }
}

static void
_dm_iter_destroy(struct dm_iter **iterp)
{
    int i;
    struct dm_iter *iter = *iterp;

    if (!iter)
        return;

    for (i = 0; i < iter->nobjs; i++) {
        struct dm_iter_object *obj;

        obj = iter->objects[i];
        if (obj->name) {
            free((char *)obj->name);
            obj->name = NULL;
        }
        if (obj->type) {
            free(obj->type);
            obj->type = NULL;
        }
        if (obj->deps) {
            free(obj->deps);
            obj->deps = NULL;
        }
        free(obj);
    }

    free(iter->names);
    free(iter);
    *iterp = NULL;
}

#define dm_iter_destroy(_i) _dm_iter_destroy(&(_i))

static struct dm_iter_object *
_dm_iter_next(struct dm_iter *iter, struct dm_iter_object *obj, int descend)
{
    int i;
    if (obj->visited)
        return NULL;
    if (obj->ndeps == 0) {
        obj->visited = 1;
        return obj;
    }
    if (descend) {
        for (i = 0; i < obj->ndeps; i++) {
            struct dm_iter_object *dep;

            dep = obj->deps[i];

            if (dep->visited)
                continue;
            return _dm_iter_next(iter, dep, 0);
        }
    }
    obj->visited = 1;
    return obj;
}

static struct dm_iter_object *
dm_iter_next(struct dm_iter *iter, int descend)
{
    int i, j;

    if (iter->prune_names) {
        for (i = 0; iter->prune_names[i]; i++) {
            struct dm_iter_object dummy = {
                .name = iter->prune_names[i],
            };
            struct dm_iter_object *dummyp = &dummy;
            struct dm_iter_object **objp, *obj;

            objp = lfind(&dummyp, iter->objects, &(iter->nobjs),
                    sizeof (struct dm_iter_object *), dm_iter_object_namesort);
            if (!objp)
                continue;
            obj = _dm_iter_next(iter, *objp, descend);
            if (obj)
                return obj;
        }
        return NULL;
    }
    for (i = 0; i < iter->nobjs; i++) {
        struct dm_iter_object *obj = iter->objects[i];
        int visit = 1;

        if (obj->visited)
            continue;
        if (obj->ndeps == 0) {
            obj->visited = 1;
            return obj;
        }
        if (descend) {
            for (j = 0; j < obj->ndeps; j++) {
                struct dm_iter_object *dep = obj->deps[j];

                if (!dep->visited) {
                    visit = 0;
                    break;
                }
            }
        }
        if (visit) {
            obj->visited = 1;
            return obj;
        }
    }
    return NULL;
}

static void
dm_print_rmparts(nashContext *nc, const char *name)
{
    char *dm = "device-mapper";
    struct proc_dev_info const *info;
    dev_t devno = makedev(-1,-1);
    struct dm_task *task;
    struct dm_deps *deps;
    int ret, i;

    i = -1;
    info = NULL;
    if ((i = getDevsFromProc(nc, i, S_IFBLK, &dm, &devno, &info)) < 0)
        return;

    ret = nashDmTaskNew(DM_DEVICE_DEPS, name, &task);
    if (ret < 0)
        return;

    deps = dm_task_get_deps(task);
    if (!deps) {
        dm_task_destroy(task);
        return;
    }

    for (i=0; i < deps->count; i++) {
        if (major(deps->device[i]) != major(devno)) {
            char *path = nashFindDeviceByDevno(nc, deps->device[i]);
            if (path)
                printf("rmparts %s\n", path);
        }
    }
    dm_task_destroy(task);
}

static int
open_part(nashContext *nc, const char *name, PedDevice **dev, PedDisk **disk)
{
    int open = 0;
    char *path = NULL;
    PedExceptionHandler *old_handler;
    int display = nashPartedErrorDisplay;

    asprintf(&path, "/dev/mapper/%s", name);

    old_handler = ped_exception_get_handler();
    nash_parted_context = nc;
    ped_exception_set_handler(nashPartedExceptionHandler);
    nashPartedErrorDisplay = 0;

    *dev = ped_device_get(path);
    free(path);
    if (!*dev || nashPartedError)
        goto out;

    if (!ped_device_open(*dev))
        goto out;
    open = 1;

    *disk = ped_disk_new(*dev);
    if (!*disk || nashPartedError)
        goto out;

    ped_exception_set_handler(old_handler);
    nashPartedError = 0;
    return 0;
out:
    if (*disk)
        ped_disk_destroy(*disk);
    *disk = NULL;
    if (open)
        ped_device_close(*dev);
    *dev = NULL;
    ped_exception_set_handler(old_handler);
    nashPartedErrorDisplay = display;
    nashPartedError = 0;
    return -1;
}

static int
dm_submap_has_part(nashContext *nc, const struct dm_iter_object *parent,
        PedGeometry *geom)
{
    int nonlinear = 0;
    int haspart = 0;
    int display = nashPartedErrorDisplay;

    /* this is a bit of a hack */
    struct dm_tree *tree;
    struct dm_tree_node *pnode, *cnode;
    void *handle = NULL;
    struct dm_iter *iter;
    struct dm_iter_object *obj;

    tree = dm_tree_create();

    iter = dm_iter_begin(NULL);
    if (!iter)
        return 0;
    while ((obj = dm_iter_next(iter, 1)))
        dm_tree_add_dev(tree, major(obj->devno), minor(obj->devno));

    pnode = dm_tree_find_node(tree, major(parent->devno), minor(parent->devno));
    
    while ((cnode = dm_tree_next_child(&handle, pnode, 1))) {
        void *next = NULL;
        struct dm_task *task;
        int ret;

        //printf("testing %s\n", dm_tree_node_get_name(cnode));
        ret = nashDmTaskNew(DM_DEVICE_TABLE, dm_tree_node_get_name(cnode),
                &task);
        if (ret < 0)
            continue;

        do {
            uint64_t start, length;
            char *type, *params;
            const char *name = NULL;
            PedDevice *dev = NULL;
            PedDisk *disk = NULL;
            PedPartition *part = NULL;

            next = dm_get_next_target(task, next, &start, &length,
                    &type, &params);
            if (!type || !params)
                continue;
            if (strcmp(type, "linear")) {
                nonlinear++;
                continue;
            }

            name = dm_tree_node_get_name(cnode);
            if (!name)
                continue;
            if (open_part(nc, name, &dev, &disk) < 0)
                continue;

            part = ped_disk_next_partition(disk, NULL);
            if (part && !nashPartedError) {
                while (part)  {
                    part = ped_disk_next_partition(disk, part);
                    if (!part || nashPartedError)
                        continue;
                    if (!ped_partition_is_active(part))
                        continue;
                    if (geom->start == part->geom.start &&
                            geom->end == part->geom.end)
                        haspart = 1;
                }
            }

            if (disk) {
                ped_disk_destroy(disk);
                ped_device_close(dev);
            }
        } while (next);
    }
    dm_tree_free(tree);

    nashPartedErrorDisplay = display;
    nashPartedError = 0;

    if (nonlinear || haspart)
        return 1;
    return 0;
}

static int
dm_should_partition(nashContext *nc, const struct dm_iter_object *obj)
{
    PedDevice *dev = NULL;
    PedDisk *disk = NULL;
    PedPartition *part = NULL;
    PedExceptionHandler *old_handler;
    int ret = 0;
    int display = nashPartedErrorDisplay;

    old_handler = ped_exception_get_handler();
    nash_parted_context = nc;
    ped_exception_set_handler(nashPartedExceptionHandler);
    nashPartedErrorDisplay = 0;

    if (open_part(nc, obj->name, &dev, &disk) < 0)
        goto out;

    part = ped_disk_next_partition(disk, NULL);
    if (!part || nashPartedError)
        goto out;

    ret = 1;
    while (part) {
        part = ped_disk_next_partition(disk, part);
        if (!part || nashPartedError)
            continue;
        if (!ped_partition_is_active(part))
            continue;
        if (dm_submap_has_part(nc, obj, &part->geom))
            ret = 0;
    }
out:
    if (disk) {
        ped_disk_destroy(disk);
        ped_device_close(dev);
    }
    ped_exception_set_handler(old_handler);
    nashPartedErrorDisplay = display;
    nashPartedError = 0;
    return ret;
}

int
nashDmListSorted(nashContext *nc, const char **names)
{
    struct dm_iter *iter;
    struct dm_iter_object *obj;
    char **newnames;

    iter = dm_iter_begin(names);
    if (!iter)
        return 1;

    newnames = calloc(1, sizeof (char *));
    while ((obj = dm_iter_next(iter, 1))) {
        dm_print_rmparts(nc, obj->name);
        printf("create %s\n", obj->name);
        if (dm_should_partition(nc, obj))
            printf("part %s\n", obj->name);
    }

    dm_iter_destroy(iter);
    return 0;
}

static int devno_is_dm(nashContext *nc, dev_t devno)
{
    int ret = 0, i = -1;
    char *type = NULL;
    struct proc_dev_info const *info = NULL;

    while ((i = getDevsFromProc(nc, i, S_IFBLK, &type, &devno, &info)) >= 0) {
        if (type && !strcmp(type, "device-mapper"))
            ret = 1;
        type = NULL;
        info = NULL;
    }

    return ret;
}

char *
nashDmDevGetName(nashContext *nc, dev_t devno)
{
    struct dm_iter *iter;
    struct dm_iter_object *obj;
    char *ret = NULL;

    if (!devno_is_dm(nc, devno))
        return NULL;

    if (!(iter = dm_iter_begin(NULL)))
        return NULL;

    while ((obj = dm_iter_next(iter, 1)) && (obj->devno != devno))
        ;

    if (obj)
        asprintf(&ret, "/dev/mapper/%s", obj->name);

    dm_iter_destroy(iter);
    return ret;
}

char *
nashDmDevGetType(nashContext *nc, dev_t devno)
{
    static char buf[32];
    struct dm_iter *iter;
    struct dm_iter_object *obj;

    if (!devno_is_dm(nc, devno))
        return NULL;

    if (!(iter = dm_iter_begin(NULL)))
        return NULL;

    while ((obj = dm_iter_next(iter, 1)) && (obj->devno != devno))
        ;

    if (obj && obj->type) {
        strncpy(buf, obj->type, 31);
        dm_iter_destroy(iter);
        return buf;
    }
    dm_iter_destroy(iter);
    return NULL;
}

/*
 * vim:ts=8:sw=4:sts=4:et
 */
