/*
 * nash.c
 *
 * Code to load modules, mount root, and get things going.
 *
 * Copyright 2002-2008 Red Hat, Inc.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author(s): Erik Troan <ewt@redhat.com>
 *            Jeremy Katz <katzj@redhat.com>
 *            Peter Jones <pjones@redhat.com>
 */

/* We internalize losetup, mount, raidautorun, and echo commands. Other
   commands are run from the filesystem. Comments and blank lines work as
   well, argument parsing is screwy. */

#define _GNU_SOURCE 1

#include "version.h"

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <net/if.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mount.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/un.h>
#include <sys/wait.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/reboot.h>
#include <termios.h>
#include <mntent.h>
#include <execinfo.h>
#include <sched.h>
#include <argz.h>
#include <asm/unistd.h>
#include <linux/fs.h>

#include <libdevmapper.h>

#include <selinux/selinux.h>

/* for showElfInterpCommand */
#include <gelf.h>
#include <libelf.h>
#include <nlist.h>
#include <elfutils/libebl.h>

#include <bdevid.h>

#include <nash.h>

#include "lib.h"
#include "util.h"
#include "block.h"
#include "dm.h"
#include "net.h"
#define HAVE_NFS 1
#include "sundries.h"
#include "adler32.h"

/* Need to tell loop.h what the actual dev_t type is. */
#undef dev_t
#if defined(__alpha) || (defined(__sparc__) && defined(__arch64__))
#define dev_t unsigned int
#else
#define dev_t unsigned short
#endif
#include <linux/loop.h>
#undef dev_t
#define dev_t dev_t

#define syslog klogctl

#include <linux/cdrom.h>
#define MD_MAJOR 9
#include <linux/raid/md_u.h>

#ifndef RAID_AUTORUN
#define RAID_AUTORUN           _IO (MD_MAJOR, 0x14)
#endif

#ifndef MNT_FORCE
#define MNT_FORCE 0x1
#endif

#ifndef MNT_DETACH
#define MNT_DETACH 0x2
#endif

#define PATH "/usr/bin:/bin:/sbin:/usr/sbin"
static char * env[] = {
    "PATH=" PATH,
    "LVM_SUPPRESS_FD_WARNINGS=1",
    NULL
};
static char sysPath[] = PATH;

/* ocPid is set when we're in otherCommand() */
static pid_t ocPid = -1;
static int exit_status = 0;

static int
searchPath(char *bin, char **resolved)
{
    char *fullPath = NULL;
    char *pathStart;
    char *pathEnd;
    int rc;

    if (!strchr(bin, '/')) {
        pathStart = sysPath;
        while (*pathStart) {
            char pec;
            pathEnd = strchr(pathStart, ':');

            if (!pathEnd)
                pathEnd = pathStart + strlen(pathStart);

            pec = *pathEnd;
            *pathEnd = '\0';

            rc = asprintfa(&fullPath, "%s/%s", pathStart, bin);
            if (!fullPath) {
                int errnum = errno;
                eprintf("error searching path: %m\n");
                return -errnum;
            }

            *pathEnd = pec;
            pathStart = pathEnd;
            if (*pathStart)
                pathStart++;

            if (!access(fullPath, X_OK)) {
                *resolved = strdup(fullPath);
                return 0;
            }
        }
    }

    if (!access(bin, X_OK)) {
        *resolved = strdup(bin);
        if (*resolved)
            return 0;
    }
    return -errno;
}

static char *
getArg(char * cmd, char * end, char ** arg)
{
    char quote = '\0';

    if (!cmd || cmd >= end)
        return NULL;

    while (isspace(*cmd) && cmd < end)
        cmd++;
    if (cmd >= end)
        return NULL;

    if (*cmd == '"')
        cmd++, quote = '"';
    else if (*cmd == '\'')
        cmd++, quote = '\'';

    if (quote) {
        *arg = cmd;

        /* This doesn't support \ escapes */
        while (cmd < end && *cmd != quote)
            cmd++;

        if (cmd == end) {
            eprintf("error: quote mismatch for %s\n", *arg);
            return NULL;
        }

        *cmd = '\0';
        cmd++;
    } else {
        *arg = cmd;
        while (!isspace(*cmd) && cmd < end)
            cmd++;
        *cmd = '\0';
        if (**arg == '$')
            *arg = getenv(*arg+1);
        if (*arg == NULL)
            *arg = "";
    }

    cmd++;

    while (isspace(*cmd))
        cmd++;

    return cmd;
}

/* get the contents of the kernel command line from /proc/cmdline */
static char *
getKernelCmdLine(void)
{
    int fd, i, errnum;
    static char * buf = NULL;

    if (buf)
        return buf;

    fd = open("/proc/cmdline", O_RDONLY);
    if (fd < 0) {
        eprintf("getKernelCmdLine: failed to open /proc/cmdline: %m\n");
        return NULL;
    }

    i = readFD(fd, &buf);
    errnum = errno;
    close(fd);
    if (i < 0) {
        eprintf("getKernelCmdLine: failed to read /proc/cmdline: %m\n");
        return NULL;
    }
    return buf;
}

/* get the start of a kernel arg "arg".  returns everything after it
 * (useful for things like getting the args to init=).  so if you only
 * want one arg, you need to terminate it at the n */
static char *
getKernelArg(char * arg)
{
    char * start, * cmdline;
    int len;

    cmdline = start = getKernelCmdLine();
    if (start == NULL)
        return NULL;
    while (*start) {
        if (isspace(*start)) {
            start++;
            continue;
        }
        len = strlen(arg);
        /* don't return if it's some other arg that just starts like
           this one */
        if (strncmp(start, arg, len) == 0) {
            if (start[len] == '=')
                return start + len + 1;
            if (!start[len] || isspace(start[len]))
                return start + len;
        }
        while (*++start && !isspace(*start))
            ;
    }

    return NULL;
}

static int
printKernelOpt(char * cmd, char * end)
{
    char * opt, * value;

    if (!(cmd = getArg(cmd, end, &opt))) {
        eprintf("kernelopt: argument expected\n");
        return 1;
    }
    value = getKernelArg(opt);
    if (value != NULL) {
        char *chptr = value;

        while (*chptr && !isspace(*chptr))
            chptr++;
        *chptr = '\0';

        if (strlen(value) == 0)
            printf("1\n");
        else
            printf("%s\n", value);
    }
    return 0;
}

static int
mountCommand(char * cmd, char * end)
{
    char * fsType = NULL;
    char * device, *spec;
    char * mntPoint;
    char * opts = NULL;
    int rc = 0;
    int flags = 0;
    char * newOpts;

    if (!(cmd = getArg(cmd, end, &spec))) {
        eprintf(
            "usage: mount [--ro] [-o <opts>] -t <type> <device> <mntpoint>\n");
        return 1;
    }

    while (cmd && *spec == '-') {
        if (!strcmp(spec, "--ro")) {
            flags |= MS_RDONLY;
        } else if (!strcmp(spec, "--bind")) {
            flags = MS_BIND;
            fsType = "none";
        } else if (!strcmp(spec, "--move")) {
            flags = MS_MOVE;
            fsType = "none";
        } else if (!strcmp(spec, "-o")) {
            cmd = getArg(cmd, end, &opts);
            if (!cmd) {
                eprintf("mount: -o requires arguments\n");
                return 1;
            }
        } else if (!strcmp(spec, "-t")) {
            if (!(cmd = getArg(cmd, end, &fsType))) {
                eprintf("mount: missing filesystem type\n");
                return 1;
            }
        }

        cmd = getArg(cmd, end, &spec);
    }

    if (!cmd) {
        eprintf("mount: missing device or mountpoint\n");
        return 1;
    }

    if (!(cmd = getArg(cmd, end, &mntPoint))) {
        struct mntent *mnt;
        FILE *fstab;

        fstab = fopen("/etc/fstab", "r");
        if (!fstab) {
            eprintf("mount: missing mount point\n");
            return 1;
        }
        do {
            if (!(mnt = getmntent(fstab))) {
                eprintf("mount: missing mount point\n");
                fclose(fstab);
                return 1;
            }
            if (!strcmp(mnt->mnt_dir, spec)) {
                spec = mnt->mnt_fsname;
                mntPoint = mnt->mnt_dir;

                if (!strcmp(mnt->mnt_type, "bind")) {
                    flags |= MS_BIND;
                    fsType = "none";
                } else
                    fsType = mnt->mnt_type;

                opts = mnt->mnt_opts;
                break;
            }
        } while(1);

        fclose(fstab);
    }

    if (!fsType) {
        eprintf("mount: filesystem type expected\n");
        return 1;
    }

    if (cmd && cmd < end) {
        eprintf("mount: unexpected arguments\n");
        return 1;
    }

    /* need to deal with options */
    if (opts) {
        char * end;
        char * start = opts;

        newOpts = alloca(strlen(opts) + 1);
        *newOpts = '\0';

        while (*start) {
            end = strchr(start, ',');
            if (!end) {
                end = start + strlen(start);
            } else {
                *end = '\0';
                end++;
            }

            if (!strcmp(start, "ro"))
                flags |= MS_RDONLY;
            else if (!strcmp(start, "rw"))
                flags &= ~MS_RDONLY;
            else if (!strcmp(start, "nosuid"))
                flags |= MS_NOSUID;
            else if (!strcmp(start, "suid"))
                flags &= ~MS_NOSUID;
            else if (!strcmp(start, "nodev"))
                flags |= MS_NODEV;
            else if (!strcmp(start, "dev"))
                flags &= ~MS_NODEV;
            else if (!strcmp(start, "noexec"))
                flags |= MS_NOEXEC;
            else if (!strcmp(start, "exec"))
                flags &= ~MS_NOEXEC;
            else if (!strcmp(start, "sync"))
                flags |= MS_SYNCHRONOUS;
            else if (!strcmp(start, "async"))
                flags &= ~MS_SYNCHRONOUS;
            else if (!strcmp(start, "nodiratime"))
                flags |= MS_NODIRATIME;
            else if (!strcmp(start, "diratime"))
                flags &= ~MS_NODIRATIME;
            else if (!strcmp(start, "noatime"))
                flags |= MS_NOATIME;
            else if (!strcmp(start, "atime"))
                flags &= ~MS_NOATIME;
            else if (!strcmp(start, "relatime"))
                flags |= MS_RELATIME;
            else if (!strcmp(start, "norelatime"))
                flags &= ~MS_RELATIME;
            else if (!strcmp(start, "remount"))
                flags |= MS_REMOUNT;
            else if (!strcmp(start, "bind"))
                flags |= MS_BIND;
            else if (!strcmp(start, "defaults"))
                ;
            else {
                if (*newOpts)
                    strcat(newOpts, ",");
                strcat(newOpts, start);
            }

            start = end;
        }

        opts = newOpts;
    }

    if (!strncmp(fsType, "nfs", 3)) {
        device = strdupa(spec);
    } else {
        device = nashGetPathBySpec(_nash_context, spec);
    }

    if (!device) {
        eprintf("mount: could not find filesystem '%s'\n", spec);
        return 1;
    }

    if (_nash_context->testing) {
        printf("mount %s%s%s-t '%s' '%s' '%s' (%s%s%s%s%s%s%s%s)\n",
                opts ? "-o '" : "",
                opts ? opts : "",
                opts ? "\' " : "",
                fsType, device, mntPoint,
                (flags & MS_RDONLY) ? "ro " : "",
                (flags & MS_NOSUID) ? "nosuid " : "",
                (flags & MS_NODEV) ? "nodev " : "",
                (flags & MS_NOEXEC) ? "noexec " : "",
                (flags & MS_SYNCHRONOUS) ? "sync " : "",
                (flags & MS_REMOUNT) ? "remount " : "",
                (flags & MS_NOATIME) ? "noatime " : "",
                (flags & MS_RELATIME) ? "relatime " : ""
            );
    } else {
        char *mount_opts = NULL;
        if (!strncmp(fsType, "nfs", 3)) {
            if (nfsmount(device, mntPoint, &flags, &opts, &mount_opts, 0)) {
                eprintf("nfsmount: error mounting %s on %s as %s: %m\n",
                        device, mntPoint, fsType);
                return 1;
            }
        } else {
            mount_opts = opts;
        }
        if (mount(device, mntPoint, fsType, flags, mount_opts) < 0) {
            eprintf("mount: error mounting %s on %s as %s: %m\n",
                    device, mntPoint, fsType);
            rc = 1;
        }
    }

    return rc;
}

static int
otherCommand(char * bin, char * cmd, char * end, int doFork, int killHp)
{
    char ** args;
    char ** nextArg, **tmpArg;
    int wpid;
    int status = 0;
    char * stdoutFile = NULL;
    int stdoutFd = fileno(stdout);

    args = (char **)alloca(128 * sizeof (char *));
    if (!args)
        return 1;
    memset(args, '\0', 128 * sizeof (char));
    nextArg = args;

    if (access(bin, X_OK)) {
        eprintf("failed to execute %s: %m\n", bin);
        return 1;
    }

    *nextArg = strdupa(bin);

    while (cmd && cmd < end) {
        nextArg++;
        cmd = getArg(cmd, end, nextArg);
    }

    if (cmd)
        nextArg++;
    *nextArg = NULL;

    /* if the next-to-last arg is a >, redirect the output properly */
    tmpArg = nextArg - 2;
    if (((nextArg - args) >= 2) && (!strcmp(*tmpArg, ">") ||
                                    !strcmp(*tmpArg, ">>"))) {
        int flags = O_CREAT | O_RDWR;
        stdoutFile = *(nextArg - 1);
        *(nextArg - 2) = NULL;

        if (*tmpArg[1] == '\0')
            flags |= O_TRUNC;
        stdoutFd = open(stdoutFile, flags, 0600);
        if (stdoutFd < 0) {
            eprintf("nash: failed to open %s: %m\n", stdoutFile);
            return 1;
        }
        lseek(stdoutFd, 0, SEEK_END);
        setFdCoe(stdoutFd, 0);
    }

    if (_nash_context->testing) {
        printf("%s ", bin);
        nextArg = args + 1;
        while (*nextArg)
            printf(" '%s'", *nextArg++);
        if (stdoutFile)
            printf(" (> %s)", stdoutFile);
        printf("\n");
    } else {
        if (!doFork || !(ocPid = fork())) {
            /* child */
            int errnum;

            dm_lib_exit(); /* ARRGH */
            if (killHp)
                nashHotplugKill(_nash_context);
            dup2(stdoutFd, 1);
            execve(args[0], args, env);
            errnum = errno; /* so we'll have it after printf */
            eprintf("ERROR: failed in exec of %s: %m\n", args[0]);
            return 1;
        }

        if (stdoutFd != fileno(stdout))
            close(stdoutFd);

        for (;;) {
            wpid = waitpid(ocPid, &status, 0);
            if (wpid == -1)
                eprintf("ERROR: Failed to wait for process %d: %m\n", wpid);

            if (!WIFEXITED(status) || WEXITSTATUS(status)) {
#if 0
                eprintf("ERROR: %s exited abnormally with value %d (pid %d)\n",
                    args[0], WEXITSTATUS(status), ocPid);
#endif
                status = WEXITSTATUS(status);
            } else {
                status = 0;
            }
            break;
        }
    }

    return status;
}

static int
lnCommand(char *cmd, char *end)
{
    char *oldpath, *newpath;
    int symbolic = 0, rc;

    if (!(cmd = getArg(cmd, end, &oldpath))) {
        eprintf("ln: argument expected\n");
        return 1;
    }

    if (!strcmp(cmd, "-s")) {
        symbolic = 1;
        if (!(cmd = getArg(cmd, end, &oldpath))) {
            eprintf("ln: argument expected\n");
            return 1;
        }
    }

    if (!(cmd = getArg(cmd, end, &newpath))) {
        eprintf("ln: argument expected\n");
        return 1;
    }

    if (symbolic)
        rc = symlink(oldpath, newpath);
    else
        rc = link(oldpath, newpath);

    if (rc > 0) {
        eprintf("ln: error: %m\n");
        return 1;
    }

    return 0;
}

static int
copyToFd(const char *file, int destfd)
{
    char * buf;
    int fd, n;

    if ((fd = open(file, O_RDONLY)) < 0) {
        eprintf("cp: error opening %s: %m\n", file);
        return 1;
    }

    buf = calloc(1024, sizeof (char));
    while ((n = read(fd, buf, 1024)) > 0) {
        write(destfd, buf, n);
    }
    close(fd);
    return 0;
}

static int
cpCommand(char *cmd, char *end)
{
    char * src;
    char * dest;
    int fd;
    int rc;

    if (!(cmd = getArg(cmd, end, &src))) {
        eprintf("cp: source argument expected\n");
        return 1;
    }

    if (!(cmd = getArg(cmd, end, &dest))) {
        eprintf("cp: destination argument expected\n");
        return 1;
    }

    if ((fd = open(dest, O_CREAT|O_TRUNC|O_WRONLY, 0600)) < 0) {
        eprintf("cp: error opening %s: %m\n", dest);
        return 1;
    }

    rc = copyToFd(src, fd);

    close(fd);

    return rc;
}

#ifdef DEBUG
static int lsdir(char *thedir, char * prefix)
{
    DIR * dir;
    struct dirent * entry;
    struct stat sb;
    char * fn;

    if (!(dir = opendir(thedir))) {
        eprintf("error opening %s: %m\n", thedir);
        return 1;
    }

    fn = calloc(1024, sizeof (char));
    while ((entry = readdir(dir))) {
        if (!strcmp(entry->d_name, ".") || !strcmp(entry->d_name, ".."))
            continue;
        snprintf(fn, 1024, "%s/%s", thedir, entry->d_name);
        stat(fn, &sb);
        printf("%s%s", prefix, fn);

        if (S_ISDIR(sb.st_mode)) {
            char * pfx;
            pfx = calloc(strlen(prefix) + 3, sizeof (char));
            sprintf(pfx, "%s  ", prefix);
            printf("/\n");
        } else if (S_ISCHR(sb.st_mode)) {
            printf(" c %d %d\n", major(sb.st_rdev), minor(sb.st_rdev));
        } else if (S_ISBLK(sb.st_mode)) {
            printf(" b %d %d\n", major(sb.st_rdev), minor(sb.st_rdev));
        } else if (S_ISLNK(sb.st_mode)) {
            char * target;
            target = calloc(1024, sizeof (char));
            readlink(fn, target, 1024);
            printf("->%s\n", target);
            free(target);
        } else {
            printf("\n");
        }
    }
    return 0;
}
#endif

static int
catCommand(char * cmd, char * end)
{
    char * file;
    int fd = fileno(stdout);

    if (!(cmd = getArg(cmd, end, &file))) {
        eprintf("cat: argument expected\n");
        return 1;
    }

    return copyToFd(file, fd);
}

#ifdef DEBUG
static int
lsCommand(char * cmd, char * end)
{
    char * dir;

    if (!(cmd = getArg(cmd, end, &dir))) {
        eprintf("ls: argument expected\n");
        return 1;
    }

    lsdir(dir, "");
    return 0;
}
#endif

static int
execCommand(char *cmd, char *end)
{
    char *bin = NULL, *fullPath = NULL;
    int killHp = 1;
    int rc;

    while ((cmd = getArg(cmd, end, &bin))) {
        if (!strcmp(bin, "--nokill")) {
            killHp = 0;
            bin = NULL;
            continue;
        }
        break;
    }
    if (!bin) {
        eprintf("exec: argument expected\n");
        return 1;
    }
    rc = searchPath(bin, &fullPath);
    if (rc < 0)
        return 1;

    rc = otherCommand(fullPath, cmd, end, 0, killHp);
    free(fullPath);
    return rc;
}

static int
exitCommand(char *cmd, char *end) {
    char *arg;
    int killHp = 1;
    int status = 0;

    while ((cmd = getArg(cmd, end, &arg)) && arg) {
        if (!strcmp(arg, "--nokill")) {
            killHp = 0;
        } else {
            status = strtol(arg, NULL, 10);
        }
    }
    if (killHp)
        nashHotplugKill(_nash_context);

    exit(status);
}

static int
losetupCommand(char * cmd, char * end)
{
    char * device;
    char * file;
    int fd;
    struct loop_info loopInfo;
    int dev;

    if (!(cmd = getArg(cmd, end, &device))) {
        eprintf("losetup: missing device\n");
        return 1;
    }

    if (!(cmd = getArg(cmd, end, &file))) {
        eprintf("losetup: missing file\n");
        return 1;
    }

    if (cmd < end) {
        eprintf("losetup: unexpected arguments\n");
        return 1;
    }

    if (_nash_context->testing) {
        printf("losetup '%s' '%s'\n", device, file);
    } else {
        dev = open(device, O_RDWR);
        if (dev < 0) {
            eprintf("losetup: failed to open %s: %m\n", device);
            return 1;
        }

        if ((fd = open64(file, O_RDWR)) < 0) {
	    if (errno == EROFS || errno == EPERM || errno == EACCES) {
	        fd = open64(file, O_RDONLY);
	    }
	    if (fd < 0) {
                eprintf("losetup: failed to open %s: %m\n", file);
                close(dev);
                return 1;
	    }
        }

        if (ioctl(dev, LOOP_SET_FD, (long)fd)) {
            eprintf("losetup: LOOP_SET_FD failed for fd %d: %m\n", fd);
            close(dev);
            close(fd);
            return 1;
        }

        close(fd);

        memset(&loopInfo, 0, sizeof(loopInfo));
        strcpy(loopInfo.lo_name, file);

        if (ioctl(dev, LOOP_SET_STATUS, &loopInfo))
            eprintf("losetup: LOOP_SET_STATUS failed: %m\n");

        close(dev);
    }

    return 0;
}

static int
hotplugCommand(char *cmd, char *end)
{
    char *new;
    cmd = getArg(cmd, end, &new);
    if (cmd) {
        eprintf("hotplug: unexpected arguments\n");
        return 1;
    }

    nashHotplugInit(_nash_context);
    return 0;
}

static int
killplugCommand(char *cmd, char *end)
{
    char *new;
    cmd = getArg(cmd, end, &new);
    if (cmd) {
        eprintf("killplug: unexpected arguments\n");
        return 1;
    }

    nashHotplugKill(_nash_context);
    return 0;
}

static int
loadDriversCommand(char *cmd, char *end)
{
    // Implements the following bash for() loop:
    // for d in /sys/bus/pci/devices/* ; do 
    //   modprobe -q $(cat $d/modalias) || : ;
    // done 

    DIR *dir;
    struct dirent *d;
    char *dirName = "/sys/bus/pci/devices";
    char strbuf[58];
    char *new;

    cmd = getArg(cmd, end, &new);
    if (cmd) {
        eprintf("loadDrivers: unexpected arguments\n");
        return 1;
    }

    if (!(dir = opendir(dirName))) {
        eprintf("loadDrivers: error opening %s: %m\n", dirName);
        return 1;
    }

    while ((d = readdir(dir))) {
        char *path = NULL;
        char *nl;
        int i;

        if (!strcmp(d->d_name, ".") || !strcmp(d->d_name, "..")) 
            continue;

        i = asprintfa(&path, "%s/%s/modalias", dirName, d->d_name);
        if (i == -1 || !path) {
            eprintf("loadDrivers: error: %m\n");
            continue;
        }

        strcpy(strbuf, "-q ");
        nl = strbuf + 3;
        i = readPathBuf(path, nl, 55);
        if (i <= 0)
            continue;
        nl += i - 1;
        if (nl[0] == '\n') {
            nl[0] = '\0';
        } else {
            nl++;
            nl[0] = '\0';
        }

        /* We don't care about the return code */
        otherCommand("/sbin/modprobe", strbuf, nl, 1, 0);
    }
    closedir(dir);
    return 0;
}

#define RAID_MAJOR 9
static int
raidautorunCommand(char * cmd, char * end)
{
    char * device;
    int fd;

    if (!(cmd = getArg(cmd, end, &device))) {
        eprintf("raidautorun: raid device expected as first argument\n");
        return 1;
    }

    if (cmd < end) {
        eprintf("raidautorun: unexpected arguments\n");
        return 1;
    }

    /* with udev, the raid devices don't exist until they get started.
     * this won't work so well with raidautorun.  so, let's be smart
     * and create them ourselves if we need to */
    if (access(device, R_OK & W_OK)) {
        int minor;
        if (sscanf(device, "/dev/md%d", &minor) != 1) {
            eprintf("raidautorun: unable to autocreate %s\n", device);
            return 1;
        }

        if (smartmknod(device, S_IFBLK | 0600, makedev(RAID_MAJOR, minor))) {
            eprintf("raidautorun: unable to autocreate %s\n", device);
            return 1;
        }
    }

    fd = open(device, O_RDWR);
    if (fd < 0) {
        eprintf("raidautorun: failed to open %s: %m\n", device);
        return 1;
    }

    if (ioctl(fd, RAID_AUTORUN, 0)) {
        eprintf("raidautorun: RAID_AUTORUN failed: %m\n");
        close(fd);
        return 1;
    }

    close(fd);
    return 0;
}

/* remove all files/directories below dirName -- don't cross mountpoints */
static int
recursiveRemove(char * dirName)
{
    struct stat sb,rb;
    DIR * dir;
    struct dirent * d;
    char * strBuf = alloca(strlen(dirName) + 1024);

    if (!(dir = opendir(dirName))) {
        eprintf("error opening %s: %m\n", dirName);
        return 0;
    }

    if (fstat(dirfd(dir),&rb)) {
        eprintf("unable to stat %s: %m\n", dirName);
        closedir(dir);
        return 0;
    }

    errno = 0;
    while ((d = readdir(dir))) {
        errno = 0;

        if (!strcmp(d->d_name, ".") || !strcmp(d->d_name, "..")) {
            errno = 0;
            continue;
        }

        strcpy(strBuf, dirName);
        strcat(strBuf, "/");
        strcat(strBuf, d->d_name);

        if (lstat(strBuf, &sb)) {
            eprintf("failed to stat %s: %m\n", strBuf);
            errno = 0;
            continue;
        }

        /* only descend into subdirectories if device is same as dir */
        if (S_ISDIR(sb.st_mode)) {
            if (sb.st_dev == rb.st_dev) {
                recursiveRemove(strBuf);
                if (rmdir(strBuf))
                    eprintf("failed to rmdir %s: %m\n", strBuf);
            }
            errno = 0;
            continue;
        }

        if (unlink(strBuf)) {
            eprintf("failed to remove %s: %m\n", strBuf);
            errno = 0;
            continue;
        }
    }

    if (errno) {
        closedir(dir);
        eprintf("error reading from %s: %m\n", dirName);
        return 1;
    }

    closedir(dir);

    return 0;
}

static void
mountMntEnt(const struct mntent *mnt)
{
    char *start = NULL, *end;
    char *target = NULL;
    struct stat sb;
    
    qprintf("mounting %s\n", mnt->mnt_dir);
    if (asprintfa(&target, ".%s", mnt->mnt_dir) < 0) {
        eprintf("setuproot: out of memory while mounting %s\n",
                mnt->mnt_dir);
        return;
    }
    
    if (stat(target, &sb) < 0)
        return;
    
    if (asprintf(&start, "-o %s -t %s %s .%s\n",
            mnt->mnt_opts, mnt->mnt_type, mnt->mnt_fsname,
            mnt->mnt_dir) < 0) {
        eprintf("setuproot: out of memory while mounting %s\n",
                mnt->mnt_dir);
        return;
    }
    
    end = start + 1;
    while (*end && (*end != '\n'))
        end++;
    /* end points to the \n at the end of the command */
    
    if (mountCommand(start, end) != 0)
        eprintf("setuproot: mount returned error\n");
}

static int
setuprootCommand(char *cmd, char *end)
{
    FILE *fp;
    char *new;

    qprintf("Setting up new root fs\n");

    cmd = getArg(cmd, end, &new);
    if (cmd) {
        eprintf("setuproot: unexpected arguments\n");
        return 1;
    }
    new = "/sysroot";

    if (chdir(new)) {
        eprintf("setuproot: chdir(%s) failed: %m\n", new);
        return 1;
    }

    if (mount("/dev", "./dev", NULL, MS_BIND, NULL) < 0)
        eprintf("setuproot: moving /dev failed: %m\n");

    if (!getKernelArg("nomnt")) {
        fp = setmntent("./etc/fstab.sys", "r");
        if (fp)
            qprintf("using fstab.sys from mounted FS\n");
        else {
            fp = setmntent("/etc/fstab.sys", "r");
            if (fp)
                qprintf("using fstab.sys from initrd\n");
        }
        if (fp) {
            struct mntent *mnt;

            while((mnt = getmntent(fp)))
                mountMntEnt(mnt);
            endmntent(fp);
        } else {
            struct {
                char *source;
                char *target;
                char *type;
                int flags;
                void *data;
                int raise;
            } fstab[] = {
                { "/proc", "./proc", "proc", 0, NULL },
                { "/sys", "./sys", "sysfs", 0, NULL },
#if 0
                { "/dev/pts", "./dev/pts", "devpts", 0, "gid=5,mode=620" },
                { "/dev/shm", "./dev/shm", "tmpfs", 0, NULL },
                { "/selinux", "/selinux", "selinuxfs", 0, NULL },
#endif
                { NULL, }
            };
            int i = 0;

            qprintf("no fstab.sys, mounting internal defaults\n");
            for (; fstab[i].source != NULL; i++) {
                if (mount(fstab[i].source, fstab[i].target, fstab[i].type,
                            fstab[i].flags, fstab[i].data) < 0)
                    eprintf("setuproot: error mounting %s: %m\n",
                            fstab[i].source);
            }
        }
    }

    chdir("/");
    return 0;
}

#define MAX_INIT_ARGS 32
static int
switchrootCommand(char * cmd, char * end)
{
    /*  Don't try to unmount the old "/", there's no way to do it. */
    const char *umounts[] = { "/dev", "/proc", "/sys", NULL };
    const char *initprogs[] = { "/sbin/init", "/etc/init",
                                "/bin/init", "/bin/sh", NULL };
    const char *ignoreargs[] = { "console=", "BOOT_IMAGE=", NULL };
    char *init, **initargs;
    char *cmdline = NULL;
    char *new = NULL, *arg;
    int fd, i = 0, j = 0;
    int do_rm = 1, do_umount = 1;

    while ((cmd = getArg(cmd, end, &arg))) {
        if (!strcmp(arg, "--no-rm")) {
            do_rm = 0;
        } else if (!strcmp(arg, "--no-umount")) {
            do_umount = 0;
        } else if (new == NULL) {
            new = arg;
        } else {
            eprintf("switchroot: unexpected arguments\n");
            return 1;
        }
    }
    if (new == NULL)
        new = "/sysroot";

    /* this has to happen before we unmount /proc */
    init = getKernelArg("init");
    if (init == NULL)
        cmdline = getKernelCmdLine();

    fd = open("/", O_RDONLY);
    for (; do_umount && umounts[i] != NULL; i++) {
        qprintf("unmounting old %s\n", umounts[i]);
        if (umount2(umounts[i], MNT_DETACH) < 0) {
            eprintf("ERROR unmounting old %s: %m\n",umounts[i]);
            eprintf("forcing unmount of %s\n", umounts[i]);
            umount2(umounts[i], MNT_FORCE);
        }
    }
    i=0;

    chdir(new);
    nashHotplugNewRoot(_nash_context);
    nashHotplugNotifyExit(_nash_context);

    if (do_rm)
        recursiveRemove("/");

    if (mount(new, "/", NULL, MS_MOVE, NULL) < 0) {
        eprintf("switchroot: mount failed: %m\n");
        close(fd);
        return 1;
    }

    if (chroot(".")) {
        eprintf("switchroot: chroot() failed: %m\n");
        close(fd);
        return 1;
    }

    /* release the old "/" */
    close(fd);

    close(3);
    if ((fd = open("/dev/console", O_RDWR)) < 0) {
        eprintf("ERROR opening /dev/console: %m\n");
        eprintf("Trying to use fd 0 instead.\n");
        fd = dup2(0, 3);
    } else {
        setFdCoe(fd, 0);
        if (fd != 3) {
            dup2(fd, 3);
            close(fd);
            fd = 3;
        }
    }
    close(0);
    dup2(fd, 0);
    close(1);
    dup2(fd, 1);
    close(2);
    dup2(fd, 2);
    close(fd);

    if (init == NULL) {
        for (i = 0; initprogs[i] != NULL; i++) {
            if (!access(initprogs[i], X_OK)) {
                init = strdup(initprogs[i]);
                break;
            }
        }
    }
    i = 0;

    initargs = (char **)calloc(MAX_INIT_ARGS+1, sizeof (char *));
    if (cmdline && init) {
        initargs[i++] = strdup(init);
    } else {
        cmdline = init;
        initargs[0] = NULL;
    }

    if (cmdline != NULL) {
        char quote ='\0';
        char * chptr, * start;

        start = chptr = cmdline;
        for (; (i < MAX_INIT_ARGS) && (*start != '\0'); i++) {
            while (*chptr && (*chptr != quote)) {
                if (isspace(*chptr) && quote == '\0')
                    break;
                if (*chptr == '"' || *chptr == '\'')
                    quote = *chptr;
                chptr++;
            }
            if (quote == '"' || quote == '\'')
                chptr++;
            if (*chptr != '\0')
                *(chptr++) = '\0';
            /*
             * There are some magic parameters added *after* everything you
             * pass, including a console= from the x86_64 kernel and
             * BOOT_IMAGE= by syslinux.  Bash doesn't know what they mean
             * so it then exits, init gets killed, disaster results.  *sigh*
             */
            for(j=0; ignoreargs[j] != NULL; j++) {
                if (cmdline == init && !strncmp(start, ignoreargs[j], strlen(ignoreargs[j]))) {
                    if (!*chptr)
                        initargs[i] = NULL;
                    else
                        i--;
                    start = chptr;
                    break;
                }
            }
            if (start == chptr) continue;
            if (strlen(start) == 0)
              i--;
            else
              initargs[i] = strdup(start);
            start = chptr;
        }
    }

    if (initargs[i-1] != NULL)
        initargs[i] = NULL;

    if (access(initargs[0], X_OK)) {
        eprintf("WARNING: can't access %s\n", initargs[0]);
    }
    dm_lib_exit(); /* ARRGH */
    nashHotplugKill(_nash_context);
    execv(initargs[0], initargs);

    eprintf("exec of init (%s) failed!!!: %m\n", initargs[0]);
    return 1;
}

static int
isEchoQuiet(int fd)
{
    if (!_nash_context->reallyquiet)
        return 0;
    if (fd != 1)
        return 0;
    return 1;
}

static int
echoCommand(char * cmd, char * end)
{
    char * args[256];
    char ** nextArg = args;
    int outFd = 1;
    int num = 0;
    int i;
    int newline = 1;
    int length = 0;
    char *string;

    if (_nash_context->testing)
        qprintf("(echo) ");

    while ((cmd = getArg(cmd, end, nextArg))) {
        if (!strncmp("-n", *nextArg, MAX(2, strlen(*nextArg)))) {
            newline = 0;
        } else {
            length += strlen(*nextArg);
            nextArg++, num++;
        }
    }
    length += num + 1;

    if ((nextArg - args >= 2) && !strcmp(*(nextArg - 2), ">")) {
        outFd = open(*(nextArg - 1), O_WRONLY | O_CREAT | O_TRUNC, 0644);
        if (outFd < 0) {
            eprintf("echo: cannot open %s for write: %m\n", *(nextArg - 1));
            return 1;
        }

        newline = 0;
        num -= 2;
    }
    string = (char *)calloc(length, sizeof (char));
    *string = '\0';
    for (i = 0; i < num;i ++) {
        if (i)
            strcat(string, " ");
        strcat(string, args[i]);
    }

    if (newline)
        strcat(string, "\n");
    if (!isEchoQuiet(outFd))
        write(outFd, string, strlen(string));

    if (outFd != 1)
        close(outFd);
    free(string);

    return 0;
}

static int
umountCommand(char * cmd, char * end)
{
    char * path;

    if (!(cmd = getArg(cmd, end, &path))) {
        eprintf("umount: path expected\n");
        return 1;
    }

    if (cmd < end) {
        eprintf("umount: unexpected arguments\n");
        return 1;
    }

    if (umount2(path,MNT_DETACH) < 0) {
        eprintf("umount %s failed: %m\n", path);
        return 1;
    }

    return 0;
}

static int
resolveDeviceCommand(char *cmd, char *end)
{
    char *spec = NULL;
    char *device = NULL;

    if (!(cmd = getArg(cmd, end, &spec))) {
        eprintf("resolveDevice: device spec expected\n");
        return 1;
    }
    if (cmd < end) {
        eprintf("resolveDevice: unexpected arguments\n");
        return 1;
    }

    device = nashGetPathBySpec(_nash_context, spec);
    if (device) {
        printf("%s\n", device);
        return 0;
    }
    return 1;
}

static int
buildEnvCommand(char *cmd, char *end)
{
    char *name = NULL;
    char *arg = NULL;
    char *value = NULL;
    ssize_t len = 1;

    if (!(cmd = getArg(cmd, end, &name))) {
        eprintf("buildEnv: environment variable expected\n");
        return 1;
    }

    value = calloc(1, len);
    if (!value) {
        eprintf("buildEnv: %m\n");
        return 1;
    }

    while ((cmd = getArg(cmd, end, &arg))) {
        len += strlen(" ") + strlen(arg);
        char *oldvalue = value;
        value = realloc(oldvalue, len);
        if (!value) {
            eprintf("buildEnv: %m\n");
            free(oldvalue);
            return 1;
        }
        if (value[0] != '\0')
            strcat(value, " ");
        strcat(value, arg);
    }
    if (value[0] == '\0') {
        eprintf("buildEnv: value expected\n");
        free(value);
        return 1;
    }
    setenv(name, value, 1);
    free(value);
    return 0;
}

static int
setDeviceEnvCommand(char *cmd, char *end)
{
    char *env = NULL;
    char *spec = NULL;
    char *device = NULL;

    if (!(cmd = getArg(cmd, end, &env))) {
        eprintf("setDeviceEnv: environment variable expected\n");
        return 1;
    }
    if (!(cmd = getArg(cmd, end, &spec))) {
        eprintf("setDeviceEnv: device spec expected\n");
        return 1;
    }
    if (cmd < end) {
        eprintf("setDeviceEnv: unexpected arguments\n");
        return 1;
    }

    device = nashGetPathBySpec(_nash_context, spec);
    if (device) {
        setenv(env, device, 1);
        return 0;
    }
    return 1;
}

/* 2.6 magic swsusp stuff */
static int
resumeCommand(char * cmd, char * end)
{
    char * resumedev = NULL;
    char * resume = NULL;
    int fd, n;
    struct stat sb;
    char buf[25];

    if (!(cmd = getArg(cmd, end, &resume))) {
        eprintf("resume: resume device expected\n");
        return 1;
    }

    if (access("/sys/power/resume", W_OK)) {
        /* eprintf("/sys/power/resume doesn't exist, can't resume!\n");*/
        return 0;
    }

    if (strstr(getKernelCmdLine(), "noresume")) {
        qprintf("noresume passed, not resuming...\n");
        return 0;
    }

    resumedev = getKernelArg("resume");
    if (resumedev == NULL) {
        resumedev = resume;
    }
    n = strcspn(resumedev, " \t\r\n");
    if (!(resumedev = strndupa(resumedev, n))) {
        eprintf("Unable to find resume device: %m\n");
        return 1;
    }

    qprintf("Trying to resume from %s\n", resumedev);

    resume = nashGetPathBySpec(_nash_context, resumedev);
    if (resume)
        resumedev = resume;

    if (access(resumedev, R_OK)) {
        eprintf("Unable to access resume device (%s)\n", resumedev);
        return 1;
    }

    fd = open(resumedev, O_RDONLY);
    if (fd < 0)
        return 1;
    if (lseek(fd, getpagesize() - 10, SEEK_SET) != getpagesize() - 10) {
        close(fd);
        return 1;
    }
    if (read(fd, &buf, 6) != 6) {
        close(fd);
        return 1;
    }
    if (strncmp(buf, "S1SUSP", 6) && strncmp(buf, "S2SUSP", 6)) {
        qprintf("No suspend signature on swap, not resuming.\n");
        close(fd);
        return 0;
    }

    if (fstat(fd, &sb)) {
        close(fd);
        return 1;
    }
    close(fd);

    printf("Resuming from %s.\n", resumedev);
    fflush(stdout);
    fd = open("/sys/power/resume", O_WRONLY);
    memset(buf, '\0', 20);
    snprintf(buf, 20, "%d:%d", major(sb.st_rdev), minor(sb.st_rdev));
    write(fd, buf, 20);
    close(fd);

    eprintf("Resume failed.  Continuing with normal startup.\n");
    return 0;
}

static inline char *
_get_dhcp_root_path(char **bufp)
{
    int fd = -1;
    int i, errnum;
    char *s = NULL, *path = NULL;

    fd = open("/var/lib/dhclient/dhclient.leases", O_RDONLY);
    if (fd < 0)
        goto out;

    i = readFD(fd, &s);
    save_errno(errnum, close(fd));
    *bufp = s;
    if (i <= 0) {
        eprintf("Failed to read dhcp lease file.\n");
        goto out;
    }

    s = strstr(s, "option root-path ");
#if 0
    /* XXX PJ FIXME: Can't do this right until gcc gets %m[ . %a[ causes a
     * warning that's not supressable in --std=gnu99 mode.  rhbz# 249723 */
    if (!s || sscanf(s, " \"%a[^\"]\"", &path) < 1)
        eprintf("Could not find root-path dhcp option.\n");
#else
    if (!s) {
        eprintf("Could not find root-path dhcp option.\n");
        goto out;
    }
    s = strchr(s, '"');
    path = s+1;
    if (!s || !*path) {
        eprintf("Could not find root-path dhcp option.\n");
        goto out;
    }
    s = strchr(path, '"');
    if (!s) {
        eprintf("Could not find root-path dhcp option.\n");
        goto out;
    }
    s[0] = '\0';
#endif

out:
    return path;
}

#define get_dhcp_root_path()                                            \
    ({                                                                  \
        char *__buf = NULL, *__path = NULL;                             \
        int __errnum;                                                   \
        __path = _get_dhcp_root_path(&__buf);                           \
        save_errno(__errnum, {                                          \
                if (__path)                                             \
                    __path = strdupa(__path);                           \
                if (__buf)                                              \
                    free(__buf);                                        \
            } );                                                        \
        __path;                                                         \
    })

static int
mkrootdevCommand(char *cmd, char *end)
{
    char *chptr = NULL, *root;
    int i;
    FILE *fstab;
    struct mntent mnt = {
        .mnt_fsname = "/dev/root",
        .mnt_dir = "/sysroot",
        .mnt_type = NULL,
        .mnt_opts = NULL,
        .mnt_freq = 0,
        .mnt_passno = 0
    };

    root = getKernelArg("root");
    if (root) {
        char c;
        chptr = root;
        while (*chptr && !isspace(*chptr))
            chptr++;
        c = *chptr;
        *chptr = '\0';
        root = strdupa(root);
        *chptr = c;
        chptr = NULL;
    }

    i = 0;
    while ((cmd = getArg(cmd, end, &chptr))) {
        if (!strcmp(chptr, "-t")) {
            cmd = getArg(cmd, end, &mnt.mnt_type);
            if (!cmd) {
                eprintf("mkrootdev: expected fs type\n");
                return 1;
            }
        } else if (!strcmp(chptr, "-o")) {
            cmd = getArg(cmd, end, &mnt.mnt_opts);
            if (!cmd) {
                eprintf("mkrootdev: expected device name\n");
                return 1;
            }
        } else if (root) {
            if (i) {
                eprintf("mkrootdev: unexpected arguments\n");
                eprintf("cmd: %p end: %p\n", cmd, end);
                eprintf("cmd: %s\n", cmd);
                return 1;
            }
            /* if we get here, we got root from the kernel command line,
               so we don't _really_ care that there wasn't one on the
               mkrootdev command line. */
            i++;
        } else {
            root = chptr;
            i++;
        }
    }

    if (!root) {
        eprintf("mkrootdev: expected device name\n");
        return 1;
    }

    if (!mnt.mnt_type) {
        eprintf("mkrootdev: expected fs type\n");
        return 1;
    }
    if (!mnt.mnt_opts) {
        eprintf("mkrootdev: expected fs options\n");
        return 1;
    }
    /* Handle "dhcp" autodetection of mnt_fsname, mnt_type, mnt_opts */
    if (!strncmp(mnt.mnt_type, "dhcp", 4)) {
        char *rootpath;
        rootpath = get_dhcp_root_path();
        if (rootpath == NULL) {
            /* error message was already printed, just return */
            return 1;
        }
        if (!strncmp(rootpath, "nbd:", 4)) {
            /* Parse NBD options and setup /dev/nbd0 */
            char *s, *ipaddr, *port, *fstype, *rootopts, *args;
            int offset;

            /* Skip nbd: */
            s = rootpath + 4;

            /* IP Address */
            ipaddr = s;
            s = strchr(s,':');
            if (s)
                *s = '\0';
            if (!s || strlen(ipaddr) > 15 || strlen(ipaddr) == 0) {
                eprintf("mkrootdev nbd: Invalid IPv4 address from DHCP root-path: %s\n", ipaddr);
                eprintf("EXPECTED   root-path \"IPADDRESS:PORT:FILESYSTEM:OPTIONS\"\n");
                eprintf("ACTUAL     root-path \"nbd:%s\"\n", ipaddr);
                return 1;
            }

            /* Port Number */
            port = ++s;
            s = strchr(s,':');
            if (s)
                *s = '\0';
            if (!s || strlen(port) > 5 || strlen(port) == 0) {
                eprintf("mkrootdev nbd: Invalid TCP port number from DHCP root-path: %s\n", port);
                eprintf("EXPECTED   root-path \"IPADDRESS:PORT:FILESYSTEM:OPTIONS\"\n");
                eprintf("ACTUAL     root-path \"nbd:%s:%s\"\n", ipaddr, port);
                return 1;
            }

            /* Filesystem Type */
            fstype = ++s;
            s = strchr(s,':');
            if (s)
                *s = '\0';
            if (!s || strlen(fstype) == 0) {
                eprintf("mkrootdev nbd: Invalid Filesystem Type from DHCP root-path: %s\n",fstype);
                eprintf("EXPECTED   root-path \"IPADDRESS:PORT:FILESYSTEM:OPTIONS\"\n");
                eprintf("ACTUAL     root-path \"nbd:%s:%s:%s\"\n", ipaddr, port,
                        fstype);
                return 1;
            }

            /* Mount Options (may be empty) */
            rootopts = ++s;
            if (rootopts[0] == 0) {
                /* Empty, set default to ro */
                rootopts = strdupa("ro");
            }

            /* Use Parameters with nbd-client to setup /dev/nbd0 for rootdev */
            /* NOTE: IPv4 addresses only */
            offset = asprintf(&args,"%s %s /dev/nbd0", ipaddr, port);
            otherCommand("/sbin/nbd-client", args, args+offset, 1, 0);
            /* Setup rootdev for later mount /sysroot */
            mnt.mnt_type = strdupa(fstype);
            mnt.mnt_fsname = strdupa("/dev/nbd0");
            mnt.mnt_opts = strdupa(rootopts);
        } else {
            /* Let the next part handle nfs root */
            mnt.mnt_type = strdupa("nfs");
        }
    }
    /* nfs can't use /dev/root */
    if (!strncmp(mnt.mnt_type, "nfs", 3)) {
        if (hasmntopt(&mnt, "dhcp"))
            mnt.mnt_fsname = get_dhcp_root_path();
        else
            mnt.mnt_fsname = strdupa(root);
        
        if (!mnt.mnt_fsname) {
            eprintf("mkrootdev: could not determine nfs root target.\n");
            return 1;
        }
    }

    umask(0122);
    fstab = fopen("/etc/fstab", "w+");
    if (!fstab) {
        eprintf("mkrootdev: could not create fstab: %m\n");
        return 1;
    }
    addmntent(fstab, &mnt);
    fclose(fstab);

    if (!strncmp(mnt.mnt_type, "nfs", 3))
        return 0;
    return nashMkPathBySpec(_nash_context, root, "/dev/root") < 0 ? 1 : 0;
}

static int
mkdirCommand(char * cmd, char * end)
{
    char * dir;
    int ignoreExists = 0;

    cmd = getArg(cmd, end, &dir);

    if (cmd && !strcmp(dir, "-p")) {
        ignoreExists = 1;
        cmd = getArg(cmd, end, &dir);
    }

    if (!cmd) {
        eprintf("mkdir: directory expected\n");
        return 1;
    }

    if (mkdir(dir, 0755)) {
        if (!ignoreExists && errno == EEXIST) {
            eprintf("mkdir: failed to create %s: %m\n", dir);
            return 1;
        }
    }

    return 0;
}


static int
loadPolicyCommand(char * cmd, char *end)
{
    int enforce = 0;
    int rootfd;

    rootfd = open("/", O_DIRECTORY|O_RDONLY);

    if (rootfd < 0) {
        eprintf("loadpolicy: could not open directory: %m\n");
        return 1;
    }
    
    if (chdir("/sysroot") != 0) {
        eprintf("loadpolicy: unable to change directory: %m\n");
        close(rootfd);
        return 1;
    }

    if (chroot("/sysroot") != 0) {
        eprintf("loadpolicy: chroot failed: %m\n");
        close(rootfd);
        return 1;
    }

    if (selinux_init_load_policy(&enforce) != 0 && enforce > 0) {
        eprintf("Unable to load SELinux policy (%m). Halting now.\n");
        sleep(-1);
        exit(1);
    }

    if (fchdir(rootfd) < 0) {
        eprintf("loadpolicy: unable to change directory: %m\n");
        sleep(-1);
        exit(1);
    }

    if (chroot("..") < 0) {
        eprintf("loadpolicy: chroot failed: %m\n");
        sleep(-1);
        exit(1);
    }

    close(rootfd);
    return 0;
}

static int
accessCommand(char * cmd, char * end)
{
    char * permStr;
    int perms = 0;
    char * file = NULL;

    cmd = getArg(cmd, end, &permStr);
    if (cmd)
        cmd = getArg(cmd, end, &file);

    if (!cmd || *permStr != '-') {
        eprintf("usage: access -[perm] file\n");
        return 1;
    }

    permStr++;
    while (*permStr) {
        switch (*permStr) {
            case 'r': perms |= R_OK; break;
            case 'w': perms |= W_OK; break;
            case 'x': perms |= X_OK; break;
            case 'f': perms |= F_OK; break;
            default:
                eprintf("perms must be -[r][w][x][f]\n");
                return 1;
        }

        permStr++;
    }

    if ((file == NULL) || (access(file, perms)))
        return 1;

    return 0;
}

static void
showElfInterpUsage(FILE *file)
{
    if (file == stderr)
        eprintf("Usage: nash-showelfinterp <filename>\n");
    else
        fprintf(file, "Usage: nash-showelfinterp <filename>\n");
}

static int
showElfInterpCommand(char *cmd, char *end)
{
    int fd;
    Elf *elf;
    GElf_Ehdr *ehdr, ehdr_mem;
    char *interp = NULL, *file = NULL;

    cmd = getArg(cmd, end, &file);
    if (cmd) {
        if (!strcmp(file, "--help")) {
            showElfInterpUsage(stdout);
            return 0;
        }

        cmd = getArg(cmd, end, &file);
    }

    if (cmd) {
        eprintf("showelfinterp: unexpected-arguments\n");
        showElfInterpUsage(stderr);
        return 1;
    }

    if (!(fd = open(file, O_RDONLY))) {
        eprintf("showelfinterp: could not open %s: %m\n", file);
        return 1;
    }

    elf_version(EV_CURRENT);

    if ((elf = elf_begin(fd, ELF_C_READ_MMAP, NULL)) == NULL) {
        printf("%s\n", file);
        close(fd);
        return 0;
    }

    if (elf_kind(elf) != ELF_K_ELF) {
        printf("%s\n", file);
        elf_end(elf);
        close(fd);
        return 0;
    }

    ehdr = gelf_getehdr(elf, &ehdr_mem);

    if (ehdr->e_type == ET_EXEC || ehdr->e_type == ET_DYN) {
        GElf_Phdr *phdr, phdr_mem;
        int i = 0;

        do {
            phdr = gelf_getphdr(elf, i, &phdr_mem);
            if (phdr == NULL)
                break;

            if (phdr->p_type == PT_INTERP) {
                interp = alloca(phdr->p_filesz+1);
                int rc = 0;

                lseek(fd, phdr->p_offset, SEEK_SET);
                while (rc < phdr->p_filesz)
                    rc += read(fd, interp + rc, phdr->p_filesz - rc);

                if (interp[0])
                    break;
            }

            i++;
        } while (phdr != 0);
        if (!interp || !interp[0])
            interp = file;
        printf("%s\n", interp);
    } else {
        /* this isn't dynamically linked, just return the input */
        printf("%s\n", file);
    }

    elf_end(elf);
    close(fd);
    return 0;
}

static int
showLabelsCommand(char *cmd, char *end)
{
    char *removable = NULL;
    int showRemovable = 0;

    cmd = getArg(cmd, end, &removable);
    if (cmd && (!strcmp(removable, "-r") || !strcmp(removable, "--removable"))) {
        showRemovable = 1;
        cmd = getArg(cmd, end, &removable);
    }

    if (cmd) {
        eprintf("showlabels: unexpected arguments\n");
        return 1;
    }
    block_show_labels(_nash_context, showRemovable);
    return 0;
}

static int
sleepCommand(char * cmd, char * end)
{
    char *delaystr;
    struct timespec delay = { 0, 0 };

    if (!(cmd = getArg(cmd, end, &delaystr))) {
        eprintf("sleep: delay expected\n");
        return 1;
    }

    delay.tv_sec = strtoll(delaystr, NULL, 0);
    udelayspec(delay);

    return 0;
}

static int
stabilizedMtime(char *path, int iterations, struct timespec interval, int goal)
{
    struct stat sb;
    struct timespec last = {0, 0};
    int count = 0, changed = 0;


    do {
        struct timeval now;
        memset(&now, '\0', sizeof(now));
        gettimeofday(&now, NULL);
        memset(&sb, '\0', sizeof(sb));
        if (stat(path, &sb) == -1) {
            eprintf("stabilized: stat %s: %m\n", path);
            return -1;
        }
#if 0
        qprintf("last: %ld.%ld sb: %ld.%ld now: %ld.%ld\n", last.tv_sec, last.tv_nsec, sb.st_mtime, sb.st_mtim.tv_nsec, now.tv_sec, now.tv_usec * 1000);
#endif
        if (sb.st_mtime == last.tv_sec
                && sb.st_mtim.tv_nsec == last.tv_nsec) {
            if (++count == goal)
                return changed;
        } else {
            changed = 1;
            count = 0;
        }

        udelayspec(interval);
        last.tv_sec = sb.st_mtim.tv_sec;
        last.tv_nsec = sb.st_mtim.tv_nsec;
        if (iterations != -1)
            iterations--;
    } while (iterations == -1 || iterations > 0);
    return -1;
}

static int
stabilizedPoll(char *path, int iterations, struct timespec interval, int goal)
{
    int count = -1, changed = 0;
    struct pollfd pd = {
        .events = POLLIN | POLLPRI | POLLERR |POLLHUP | POLLMSG,
        .revents = 0,
    };
    struct timespec timeout;
    char buf;
    int last = -1, rc = 0;

    if ((pd.fd = open(path, O_RDONLY|O_NONBLOCK)) < 0) {
        eprintf("stabilized: open %s: %m\n", path);
        return -1;
    }

    /* first, drain the file */
    while ((rc = read(pd.fd, &buf, 1)) >= 0) {
        if (rc == 0 && last == 0)
            break;
        last = rc;
    }
    timeout.tv_sec = 0;
    timeout.tv_nsec = 1;
    rc = nash_ppoll(&pd, 1, &timeout, NULL, 0);

    struct timespec *top = specinf(&interval) ? NULL : &timeout;
    do {
        timeout = interval;

        pd.revents = 0;
        while ((rc = nash_ppoll(&pd, 1, top, NULL, 0) < 0)) {
            if (errno != EINTR)
                return -1;
        }
#if 0
        qprintf("rc: %d pd.revent: %d\n", rc, pd.revents);
#endif
        if (pd.revents) {
            changed = 1;
            count = 0;
        } else {
            if (++count == goal)
                return changed;
        }
        if (iterations != -1)
            iterations--;
    } while (iterations == -1 || iterations > 0);
    return -1;
}

static int
stabilizedHash(char *path, int iterations, struct timespec interval, int goal)
{
    int fd;
    uint32_t last = -1;
    int count = 0, changed = 0;
    int ret = -1;

    if ((fd = open(path, O_RDONLY)) < 0) {
        eprintf("stabilized: open %s: %m\n", path);
        return -1;
    }

    do {
        uint32_t a32 = checksumFd(fd);

        if (a32 == last) {
            if (++count == goal) {
                ret = changed;
                break;
            }
        } else if (last != -1) {
            changed = 1;
            count = 0;
        }
        last = a32;

        udelayspec(interval);

        if (iterations != -1)
            iterations--;
    } while (iterations == -1 || iterations > 0);

    close(fd);
    return ret;
}

static int
stabilizedCommand(char *cmd, char *end)
{
    struct timespec interval_ts = {0,0};
    int iterations=-1;
    long long interval=750;
    char *buf = NULL, *file = NULL;
    int rc, do_poll = 0, do_hash = 0;

    while ((cmd = getArg(cmd, end, &buf))) {
        if (!strcmp(buf, "--poll")) {
            do_poll = 1;
            continue;
        } else if (!strcmp(buf, "--hash")) {
            do_hash = 1;
            continue;
        }
        if (!strcmp(buf, "--iterations")) {
            if (!(cmd = getArg(cmd, end, &buf))) {
usage:
                eprintf("usage: stabilized [--poll|--hash] [ --iterations N ] "
                        "[ --interval MILLISECS ] <file>\n");
                return 1;
            }
            iterations = strtoll(buf, NULL, 0);
            continue;
        }
        if (!strcmp(buf, "--interval")) {
            if (!(cmd = getArg(cmd, end, &buf)))
                goto usage;
                interval = strtoll(buf, NULL, 0);
            continue;
        }
        if (file)
            goto usage;
        file = buf;
    }

    if (!file)
        goto usage;

    interval_ts.tv_sec = interval / 1000;
    interval_ts.tv_nsec = (interval % 1000) * 1000000;

    if (do_poll)
        rc = stabilizedPoll(file, iterations, interval_ts, 10);
    else if (do_hash)
        rc = stabilizedHash(file, iterations, interval_ts, 10);
    else
        rc = stabilizedMtime(file, iterations, interval_ts, 10);

    if (rc < 0)
        return 1;
    if (rc == 0) {
        eprintf("Could not detect stabilization, waiting 10 seconds.\n");
        sleep(10);
    }
    return 0;
}

static int
readlinkCommand(char * cmd, char * end)
{
    char * path;
    char * buf, * respath, * fullpath = NULL;
    struct stat sb;
    int rc = 0;

    if (!(cmd = getArg(cmd, end, &path))) {
        eprintf("readlink: file expected\n");
        return 1;
    }

    if (lstat(path, &sb) == -1) {
        eprintf("unable to stat %s: %m\n", path);
        return 1;
    }

    if (!S_ISLNK(sb.st_mode)) {
        printf("%s\n", path);
        return 0;
    }

    buf = readlinka(path);
    if (buf == NULL) {
        eprintf("error readlink %s: %m\n", path);
        return 1;
    }

    /* symlink is absolute */
    if (buf[0] == '/') {
        printf("%s\n", buf);
        return 0;
    }

    /* nope, need to handle the relative symlink case too */
    respath = strrchr(path, '/');
    if (respath) {
        *respath = '\0';
    }

    asprintfa(&fullpath, "%s/%s", path, buf);
    /* and normalize it */
    respath = NULL;
    respath = canonicalize_file_name(fullpath);
    if (respath == NULL) {
        eprintf("error resolving symbolic link %s: %m\n", fullpath);
        return 1;
    }

    printf("%s\n", respath);
    free(respath);
    return rc;
}

static int
doFind(char * dirName, char * name, mode_t mask, dev_t *devno)
{
    struct stat sb;
    DIR * dir;
    struct dirent * d;
    char * strBuf = alloca(strlen(dirName) + 1024);
    int maj = -1, min = -1;

    if (devno) {
        maj = major(*devno);
        min = minor(*devno);
    }

    if (!(dir = opendir(dirName))) {
        eprintf("error opening %s: %m\n", dirName);
        return 0;
    }

    /* Avoid getting // in the output */
    if (dirName[strlen(dirName) - 1] == '/')
        dirName[strlen(dirName) - 1] = 0;

    errno = 0;
    while ((d = readdir(dir))) {
        errno = 0;

        strcpy(strBuf, dirName);
        strcat(strBuf, "/");
        strcat(strBuf, d->d_name);

        if (!strcmp(d->d_name, ".") || !strcmp(d->d_name, "..")) {
            errno = 0;
            continue;
        }

        if (lstat(strBuf, &sb)) {
            eprintf("failed to stat %s: %m\n", strBuf);
            errno = 0;
            continue;
        }

        if ((!name && !devno) ||
                (name && !strcmp(d->d_name, name)) ||
                (devno && maj == major(sb.st_rdev) &&
                          min == minor(sb.st_rdev))) {
            if (mask == 0 || (sb.st_mode & mask) == mask)
                printf("%s\n", strBuf);
        }

        if (S_ISDIR(sb.st_mode))
            doFind(strBuf, name, mask, devno);
    }

    if (errno) {
        closedir(dir);
        eprintf("error reading from %s: %m\n", dirName);
        return 1;
    }

    closedir(dir);

    return 0;
}

static int
findCommand(char * cmd, char * end)
{
    char * dir;
    char * name = NULL;
    char * type = NULL;
    mode_t mode = 0;
    int maj = -1, min;
    dev_t devno = makedev(-1, -1);

    if (!(cmd = getArg(cmd, end, &dir))) {
        dir = strdupa(".");
    } else {
        if (!strcmp(dir, "-type")) {
            if (!cmd)
                goto error;
            if (!(cmd = getArg(cmd, end, &type)))
                goto error;
            if (strlen(type) != 1) {
                eprintf("find: error: invalid type '%s'\n", type);
                goto error;
            }
            switch (*type) {
                case 'c':
                    mode = S_IFCHR;
                    /* fall through */
                case 'b':
                    if (mode == 0)
                        mode = S_IFBLK;
                    if (!(cmd = getArg(cmd, end, &dir))) {
                        eprintf("find: error: expected device major:minor pair\n");
                        goto error;
                    }
                    if (sscanf(dir, "%d:%d", &maj, &min) != 2) {
                        eprintf("find: error: expected device major:minor pair\n");
                        goto error;
                    }
                    devno = makedev(maj,min);
                    /* fall through */
                case 'd':
                    if (mode == 0)
                        mode = S_IFDIR;
                    if (!(cmd = getArg(cmd, end, &dir)))
                        dir = strdup(".");
                    break;
                default:
                    eprintf("find: error: invalid type '%s'\n", type);
                    goto error;
            }
        }
        if ((cmd = getArg(cmd, end, &name))) {
            if (strcmp(name, "-name"))
                goto error;
            if (!(cmd = getArg(cmd, end, &name)))
                goto error;
        }
    }

    return doFind(dir, name, mode, maj == -1 ? NULL : &devno);
error:
    eprintf("usage: find [-type type] [path [-name file]]\n");
    return 1;
}

static int
mknodCommand(char * cmd, char * end)
{
    char * path, * type;
    char * majorStr, * minorStr;
    int major;
    int minor;
    char * chptr;
    mode_t mode;

    cmd = getArg(cmd, end, &path);
    if (!path)
        goto err;

    cmd = getArg(cmd, end, &type);
    if (!type)
        goto err;

    cmd = getArg(cmd, end, &majorStr);
    if (!majorStr)
        goto err;

    cmd = getArg(cmd, end, &minorStr);
    if (!minorStr) {
err:
        eprintf("mknod: usage mknod <path> [c|b] <major> <minor>\n");
        return 1;
    }

    if (!strcmp(type, "b")) {
        mode = S_IFBLK;
    } else if (!strcmp(type, "c")) {
        mode = S_IFCHR;
    } else {
        eprintf("mknod: invalid type\n");
        return 1;
    }

    major = strtol(majorStr, &chptr, 10);
    if (*chptr) {
        eprintf("invalid major number\n");
        return 1;
    }

    minor = strtol(minorStr, &chptr, 10);
    if (*chptr) {
        eprintf("invalid minor number\n");
        return 1;
    }

    if (smartmknod(path, mode | 0600, makedev(major, minor))) {
        eprintf("mknod: failed to create %s: %m\n", path);
        return 1;
    }

    return 0;
}

extern void print_dev_tree(nashContext *nc, int blktab);
static int
dmCommand(char *cmd, char *end)
{
    char *action = NULL;
    char *name = NULL;
    int rc;

    cmd = getArg(cmd, end, &action);
    if (!cmd)
        goto usage;

    if (!strcmp(action, "create")) {
        long long start, length;
        char *type = NULL, *params = NULL;
        char *c = NULL;
        char *uuid = NULL;
        
        cmd = getArg(cmd, end, &name);
        if (!cmd)
            goto usage;

        cmd = getArg(cmd, end, &uuid);
        if (!cmd)
            goto usage;

        if (!strcmp(uuid, "--uuid")) {
            if (!(cmd = getArg(cmd, end, &uuid))) {
                eprintf("dm create: missing uuid argument\n");
                return 1;
            }

            cmd = getArg(cmd, end, &params);
            if (!cmd)
                goto usage;
        } else {
            params = uuid;
            uuid = NULL;
        }

        errno = 0;
        start = strtoll(params, NULL, 0);
        if (errno)
            goto usage;

        cmd = getArg(cmd, end, &params);
        if (!cmd)
            goto usage;
        errno = 0;
        length = strtoll(params, NULL, 0);
        if (errno)
            goto usage;

        cmd = getArg(cmd, end, &type);
        if (!cmd)
            goto usage;

        params = cmd;
        c = strchr(params, '\n');
        if (c)
            *c = '\0';
        if (name[0] == '@') {
            if (nash_list_len(_nash_context->blktab->devs) > 0) {
                rc = nash_dev_tree_setup_device(_nash_context, name);
                if (rc == -1) {
                    eprintf("Could not create device %s\n", name);
#if 0
                    print_dev_tree(_nash_context, 0);
                    print_dev_tree(_nash_context, 1);
#endif
                    rc = 1;
                } else if (rc >= 0) {
                    if (rc == 0)
                        eprintf("Creating device %s in degraded state\n", name);
                    rc = nashDmCreate(_nash_context, name, uuid, start, length,
                            type, params);
                }
            } else {
                eprintf("Unknown device \"%s\"\n", name+1);
            }
        } else {
            rc = nashDmCreate(_nash_context, name, uuid, start, length, type,
                params);
        }
        if (c)
            *c = '\n';
        return rc;
    } else if (!strcmp(action, "remove")) {
        cmd = getArg(cmd, end, &name);
        if (!cmd)
            goto usage;

        if (nashDmRemove(name))
            return 0;
    } else if (!strcmp(action, "partadd")) {
        cmd = getArg(cmd, end, &name);
        if (!cmd)
            goto usage;

        if (nashDmCreatePartitions(_nash_context, name))
            return 0;
        return 1;
    } else if (!strcmp(action, "get_name")) {
        dev_t devno;
        long maj, min;
        char *tmp;

        cmd = getArg(cmd, end, &tmp);
        if (!cmd)
            goto usage;

        if (sscanf(tmp, "%ld:%ld", &maj, &min) != 2)
            goto usage;

        devno = makedev(maj, min);
        if (!(tmp = nashDmDevGetName(_nash_context, devno)))
            return 1;

        char *name = strrchr(tmp, '/');
        if (!name || (name && !*(++name)))
            name = tmp;

        printf("%s\n", name);
        free(tmp);
        return 0;
    } else if (!strcmp(action, "get_uuid")) {
        char *uuid;

        cmd = getArg(cmd, end, &name);
        if (!cmd)
            goto usage;

        uuid = nashDmGetUUID(name);
        if (uuid) {
            printf("%s\n", uuid);
            free(uuid);
            return 0;
        }
        return 1;
    } else if (!strcmp(action, "list")) {
        const char **names = NULL;
        int m = 0, n = 0;

        names = calloc(m+1, sizeof (char *));
        while((cmd = getArg(cmd, end, &name))) {
            names = realloc(names, sizeof (char *) * (m+2));
            names[m++] = strdupa(name);
            names[m] = NULL;
        }

        qsort(names, m, sizeof (char *), stringsort);
        /* it's too bad qsort doesn't dedupe... */
        for (n = 0; n < m-1; n++) {
            if (!strcmp(names[n], names[n+1])) {
                memmove(names+n, names+n+1, (m - n) * sizeof (char *));
                m--; n--;
            }
        }

        n = nashDmListSorted(_nash_context, names);
        if (names)
            free(names);
        if (n)
            return 0;
        return 1;
    }
usage:
    eprintf("usage: dm create name start length type PARAMS...\n"
            "       dm remove name\n"
            "       dm partadd path\n"
            "       dm get_name major:minor\n"
            "       dm get_uuid path\n"
            "       dm list [name0] [name1] [...] [nameN]\n"
            );
    return 1;
}

static int
mkDMNodCommand(char * cmd, char * end)
{
    int n = -1;
    dev_t devno = makedev(-1,-1);
    char *name = "device-mapper";
    int major, minor;
    struct proc_dev_info const *info;

    if ((n = getDevsFromProc(_nash_context, n, S_IFCHR, &name, &devno,
                    &info)) < 0) {
        eprintf("Unable to find device-mapper major/minor\n");
        return 1;
    }

    major = major(devno);
    minor = minor(devno);

    if (!access("/dev/mapper/control", R_OK)) {
        struct stat sb;
        if (stat("/dev/mapper/control", &sb) == 0) {
            if (S_ISCHR(sb.st_mode) && (sb.st_rdev == makedev(major, minor)))
                return 0;
        }

        unlink("/dev/mapper/control");
    }

    if (smartmknod("/dev/mapper/control", S_IFCHR | 0600, devno)) {
        eprintf("failed to create /dev/mapper/control\n");
        return 1;
    }

    return 0;
}

static int
mkchardevsCommand(char *cmd, char *end)
{
    DIR *dir;
    struct dirent *entry;

    if (!(dir = opendir("/sys/dev/char"))) {
#if 0 /* disabling the message for now; it's a relatively new kernel feature */
        eprintf("Failed to open /sys/dev/char: %m\n");
#endif
        return 1;
    }

    while ((entry = readdir(dir))) {
        char linkbuf[2048];
        char syspath[strlen(entry->d_name) + strlen("/sys/dev/char/") + 1];
        char *base;
        int major;
        int minor;
        ssize_t rc;

        if (!strcmp(entry->d_name, ".") || !strcmp(entry->d_name, ".."))
            continue;

        if (sscanf(entry->d_name, "%d:%d", &major, &minor) != 2)
            continue;

        sprintf(syspath, "/sys/dev/char/%s", entry->d_name);
        rc = readlink(syspath, linkbuf, 2047);
        linkbuf[rc] = '\0';

        base = basename(linkbuf);

        if (base != NULL && base[0] != '\0') {
            char devpath[strlen(base) + 6];

            sprintf(devpath, "/dev/%s", base);

            smartmknod(devpath, S_IFCHR | 0600, makedev(major, minor));
        }
    }
    return 0;
}

static int
mkblktabCommand(char *cmd, char *end)
{
    char *devname;
    int rc;

    cmd = getArg(cmd, end, &devname);
    if (!cmd) {
        eprintf("usage: mkblktab hda\n");
        return 1;
    }
    if (cmd < end) {
        eprintf("mkblktab: unexpected arguments\n");
        return 1;
    }
    if ((rc = nashWaitForDevice(_nash_context, devname, 0)) < 1) {
#if 0
        printf("rc == %d, error\n", rc);
        print_dev_tree(_nash_context, 0);
#endif
        return 1;
    }
    nash_dev_tree_print_blktab(_nash_context, devname);

    return 0;
}

static int
mkblkdevsCommand(char * cmd, char * end)
{
    struct nash_block_dev *dev = NULL;
    nashBdevIter iter;
    struct timespec timeout;

    if (cmd < end) {
        eprintf("mkblkdevs: unexpected arguments\n");
        return 1;
    }

    usectospec(15, &timeout);
    if ((iter = nashBdevIterNewPoll(_nash_context, &timeout)) == NULL)
        return 1;
    while (nashBdevIterNext(&iter, &dev) > 0)
        smartmknod(dev->dev_path, S_IFBLK | 0700, dev->devno);

    return 0;
}

static int
rmpartsCommand(char *cmd, char *end)
{
    char *devname = NULL;

    cmd = getArg(cmd, end, &devname);
    if (!cmd) {
        printf("usage: rmparts hda\n");
        return 1;
    }

    if (nashDisablePartitions(devname) < 1)
        return 1;
    return 0;
}

struct commandHandler {
    char *name;
    int (*fp)(char *cmd, char *end);
};

static const struct commandHandler *getCommandHandler(const char *name);

static int
statusCommand(char *cmd, char *end)
{
    char *si = NULL;
    long i;

    cmd = getArg(cmd, end, &si);
    if (!si) {
        printf("%d\n", exit_status);
        return exit_status;
    }

    i = strtoul(si, NULL, 0);
    if (i == ULONG_MAX && errno) {
        eprintf("status: error: %m\n");
        return 1;
    }

    return i;
}

static int
condCommand(char *cmd, char *end)
{
    char *op = NULL;
    int conditional = 0;
    int internal = 0;
    int rc;

    while((cmd = getArg(cmd, end, &op))) {
        char *stval = NULL;
        unsigned long tval;

        if (!op)
            return 1;

        if (op[0] != '-')
            break;

        cmd = getArg(cmd, end, &stval);
        if (stval)
            tval = strtoul(stval, NULL, 0);
        if (!stval || (tval == ULONG_MAX && errno)) {
            if (errno)
                eprintf("cond: error: %m\n");
            else
                eprintf("cond: error: invalid value '%s' for '%s' operator\n",
                        stval, op);
            return 1;
        }

        if (!strcmp(op, "-lt")) { 
            if (exit_status >= tval)
                return exit_status;
            conditional = 1;
        } else if (!strcmp(op, "-le")) {
            if (exit_status > tval)
                return exit_status;
            conditional = 1;
        } else if (!strcmp(op, "-eq") ||
                   !strcmp(op, "-et")) {
            if (exit_status != tval)
                return exit_status;
            conditional = 1;
        } else if (!strcmp(op, "-ne")) {
            if (exit_status == tval)
                return exit_status;
            conditional = 1;
        } else if (!strcmp(op, "-ge")) {
            if (exit_status < tval)
                return exit_status;
            conditional = 1;
        } else if (!strcmp(op, "-gt")) {
            if (exit_status <= tval)
                return exit_status;
            conditional = 1;
        }
    }

    if (!conditional && exit_status != 0)
        return exit_status;

    if (strncmp(op, "nash-", 5)) {
        char *fullPath = NULL;

        rc = searchPath(op, &fullPath);
        if (rc >= 0) {
            rc = otherCommand(fullPath, cmd, end, 1, 0);
            free(fullPath);
            return rc;
        } else
            internal = 1;
    } else {
        op += 5;
        internal = 1;
    }

    rc = 1;
    if (internal) {
        const struct commandHandler *handler;

        handler = getCommandHandler(op);
        if (handler->name != NULL)
            rc = (handler->fp)(cmd, end);
    }
    return rc;
}

static int
netnameCommand(char *cmd, char *end)
{
    char *hwaddr = NULL, *ifname = NULL;

    cmd = getArg(cmd, end, &hwaddr);
    if (!cmd) {
err_missing_args:
        eprintf("netname: missing arguments.\n");
usage:
        eprintf("usage: netname c0:ff:ee:ab:c1:23 eth0\n");
        return 1;
    }

    cmd = getArg(cmd, end, &ifname);
    if (!cmd)
        goto err_missing_args;

    cmd = getArg(cmd, end, &ifname);
    if (cmd) {
        eprintf("netname: extra arguments.\n");
        goto usage;
    }

    return renameNetworkInterface(hwaddr, ifname);
}

static int
networkCommand(char *cmd, char *end)
{
    size_t argz_len = 0;
    char *argz = NULL;

    int argc;
    char **argv;

    char *initial_argv[] = { "network", NULL };
    char *value = NULL;
    int rc;

    /* popt expects to get network --args here */
    if (!cmd || cmd >= end)
        return 1;

    argz_create(initial_argv, &argz, &argz_len);

    while ((cmd = getArg(cmd, end, &value)))
        argz_add(&argz, &argz_len, value);

    argc = argz_count(argz, argz_len);
    argv = calloc(argc + 1, sizeof (char *));

    argz_extract(argz, argz_len, argv);

    rc = nashNetworkCommand(argc, argv);

    free(argv);
    while (argz)
        argz_delete(&argz, &argz_len, argz);

    return rc;
}

static int
nullCommand(char *cmd, char *end)
{
    return 0;
}

static int
daemonizeCommand(char *cmd, char *end)
{
    char *tmp;
    char *path = NULL;
    int ret = 1;
    int ptmx;
    int pts;
    pid_t pid;
    char *ptn;


    /* PJFIX ought to introduce some way to specify arguments to the daemon */
    while ((cmd = getArg(cmd, end, &tmp))) {
        if (!strcmp(tmp, "--ignore-missing"))
            ret = 0;
        else if (!path)
            path = tmp;
        else {
            eprintf("daemonize: unexpected arguments\n");
            return 1;
        }
    }

    if (!path) {
        eprintf("usage: daemonize <path-to-daemon>\n");
        return 1;
    }

    if (access(path, X_OK) < 0) {
        /* if the daemon isn't installed, we quietly ignore it rather than
         * returning an error */
        if (ret)
            eprintf("daemonize: cannot execute \"%s\": %m\n", path);
        return ret;
    }
    
    ptmx = posix_openpt(O_RDWR | O_NOCTTY);
    if (ptmx < 0) {
        eprintf("daemonize: could not open pty master: %m\n");
        return 1;
    }

    ptn = ptsname(ptmx);
    if (ptn)
        ptn = strdupa(ptn);
    if (!ptn) {
        eprintf("daemonize: could not find pty slave device name: %m\n");
        close(ptmx);
        return 1;
    }

    if (grantpt(ptmx) < 0) {
        eprintf("daemonize: could not grant pty permissions: %m\n");
        close(ptmx);
        return 1;
    }

    if (unlockpt(ptmx) < 0) {
        eprintf("daemonize: could not unlock pty: %m\n");
        close(ptmx);
        return 1;
    }

    pid = fork();
    if (pid < 0) {
        eprintf("daemonize: could not fork: %m\n");
        close(ptmx);
        return 1;
    }
    if (pid != 0) {
        /* we're the parent process, which is going to remain being nash */
        /* first, get rid of ptmx */
        close(ptmx);

        /* now, set us up as the session leader, and set up our pts as
         * stdin/stdout/stderr and our controlling tty. */
        setsid();

        pts = open(ptn, O_RDONLY);
        if (pts < 0) {
            eprintf("daemonize: could not open pt slave device: %m\n");
            return 1;
        }

        if (pts != STDIN_FILENO) {
            close(STDIN_FILENO);
            dup2(pts, STDIN_FILENO);
            close(pts);
        }
        pts = open(ptn, O_WRONLY);
        if (pts < 0) {
            eprintf("daemonize: could not open pt slave device: %m\n");
            return 1;
        }
        
        if (pts == STDOUT_FILENO) {
            close(STDERR_FILENO);
            dup2(pts, STDERR_FILENO);
        } else if (pts == STDERR_FILENO) {
            close(STDOUT_FILENO);
            dup2(pts, STDOUT_FILENO);
        } else {
            close(STDOUT_FILENO);
            dup2(pts, STDOUT_FILENO);
            close(STDERR_FILENO);
            dup2(pts, STDERR_FILENO);
            close(pts);
        }

        while (1) {
            siginfo_t si;
            int rc;

            memset(&si, '\0', sizeof (si));

            rc = waitid(P_PID, pid, &si, WEXITED);
            if (rc == -1 && errno == EINVAL)
                continue;
            break;
        }

        return 0;
    } else {
        /* this is the parent, which is going to be plymouth */
        //char *argv0 = "/usr/libexec/plymouth/plymouthd";
        char *argv[] = { path, "--attach-to-session", NULL, NULL };

        asprintf(&argv[2], "%d", ptmx);

        if (execv(argv[0], argv) < 0) {
            eprintf("execv(%s, {%s,%s,%s}) failed: %m\n", argv[0],
                    argv[0], argv[1], argv[2]);
            exit(1);
        }
    }

    return 0;
}

static int
waitdevCommand(char *cmd, char *end)
{
    char *device = NULL;
    long long timeout = -2;
    int rc = 0;

    while (1) {
        if (!(cmd = getArg(cmd, end, &device))) {
            eprintf("waitdev: block device name expected\n");
            return 1;
        }
        if (timeout == -2 && !strncmp(device, "--timeout", 9)) {
            char *to_str = NULL;

            if (device[9] == '=') {
                to_str = &device[10];
            } else {
                if (!(cmd = getArg(cmd, end, &to_str))) {
                    eprintf("waitdev: value expected\n");
                    return 1;
                }
            }
            errno = 0;
            timeout = strtoll(to_str, NULL, 0);
            if ((timeout == LLONG_MIN || timeout == LLONG_MAX) && errno != 0){
                eprintf("waitdev: \"%s\": %m\n", to_str);
                return 1;
            }
        } else {
            break;
        }
    }

    if (timeout == -2)
        timeout = -1;

    if (cmd < end) {
        eprintf("waitdev: unexpected arguments\n");
        return 1;
    }

    rc = 1;
    if (nashWaitForDevice(_nash_context, device, timeout) > 0)
        rc = 0;

    return rc;
}

static int
setQuietCommand(char * cmd, char * end)
{
    char *quietcmd;

    cmd = getArg(cmd, end, &quietcmd);
    if (cmd && cmd < end) {
        eprintf("setquiet: unexpected arguments\n");
        return 1;
    }

    if (cmd && !strcmp(quietcmd, "--noisy")) {
        _nash_context->reallyquiet = 0;
        _nash_context->quiet = 0;
    } else if ((quietcmd = getKernelArg("noisy"))) {
        _nash_context->reallyquiet = 0;
        _nash_context->quiet = 0;
    } else {
        quietcmd = getKernelArg("quiet");
        if (quietcmd)
            _nash_context->reallyquiet = 1;

        quietcmd = getKernelArg("noquiet");
        if (quietcmd)
            _nash_context->reallyquiet = 0;


        /* reallyquiet may be set elsewhere */
        if (_nash_context->reallyquiet)
              _nash_context->quiet = 1;
    }

    return 0;
}

static const struct commandHandler handlers[] = {
    { "access", accessCommand },
    { "buildEnv", buildEnvCommand },
#ifdef DEBUG
    { "cat", catCommand },
#endif
    { "cond", condCommand },
    { "cp", cpCommand },
    { "daemonize", daemonizeCommand },
    { "dm", dmCommand },
    { "echo", echoCommand },
    { "exec", execCommand },
    { "exit", exitCommand },
    { "find", findCommand },
    { "kernelopt", printKernelOpt },
    { "loadDrivers", loadDriversCommand },
    { "loadpolicy", loadPolicyCommand },
    { "mkchardevs", mkchardevsCommand },
    { "mkblktab", mkblktabCommand },
    { "mkblkdevs", mkblkdevsCommand },
    { "mkdir", mkdirCommand },
    { "mkdmnod", mkDMNodCommand },
    { "mknod", mknodCommand },
    { "mkrootdev", mkrootdevCommand },
    { "mount", mountCommand },
    { "netname", netnameCommand },
    { "network", networkCommand },
    { "null", nullCommand },
    { "plymouth", nullCommand },
    { "hotplug", hotplugCommand },
    { "killplug", killplugCommand },
    { "losetup", losetupCommand },
    { "ln", lnCommand },
#ifdef DEBUG
    { "ls", lsCommand },
#endif
    { "raidautorun", raidautorunCommand },
    { "readlink", readlinkCommand },
    { "resume", resumeCommand },
    { "resolveDevice", resolveDeviceCommand },
    { "rmparts", rmpartsCommand },
    { "setDeviceEnv", setDeviceEnvCommand },
    { "setquiet", setQuietCommand },
    { "setuproot", setuprootCommand },
    { "showelfinterp", showElfInterpCommand },
    { "showlabels", showLabelsCommand },
    { "sleep", sleepCommand },
    { "stabilized", stabilizedCommand },
    { "status", statusCommand },
    { "switchroot", switchrootCommand },
    { "umount", umountCommand },
    { "waitdev", waitdevCommand },
    { NULL, },
};

static const struct commandHandler *
getCommandHandler(const char *name)
{
    const struct commandHandler *handler = NULL;

    for (handler = &handlers[0]; handler->name; handler++)
        if (!strcmp(name, handler->name))
            return handler;
    return handler;
}

static int
runStartup(int fd, char *name)
{
    char *contents = NULL;
    int i;
    char * start, * end;
    char * chptr;
    int rc;
    const struct commandHandler *handler;

    i = readFD(fd, &contents);
    if (i < 0) {
        eprintf("Failed to read startup file %s", name);
        return 1;
    }

    start = contents;
    while (start && *start) {
        while (isspace(*start) && *start && (*start != '\n'))
            start++;

        if (*start == '#') {
            while (*start && (*start != '\n'))
                start++;
            if (*start == '\n')
                start++;
        }

        if (*start == '\n') {
            start++;
            continue;
        }

        if (!*start) {
            eprintf("(last line in %s is empty)\n", name);
            continue;
        }

        /* start points to the beginning of the command */
        end = start + 1;
        while (*end && (*end != '\n'))
            end++;
        if (!*end) {
            eprintf("(last line in %s missing newline -- skipping)\n", name);
            start = end;
            continue;
        }

        /* end points to the \n at the end of the command */

        chptr = start;
        while (chptr < end && !isspace(*chptr))
            chptr++;

        i = 0;
        rc = 1;
        *(chptr++) = '\0';
        if (strncmp(start, "nash-", 5))  {
            char *fullPath = NULL;
            rc = searchPath(start, &fullPath);
            if (rc >= 0) {
                rc = otherCommand(fullPath, chptr, end, 1, 0);
                free(fullPath);
            } else
                i = 1;
        } else {
            start += 5;
            i = 1;
        }

        if (i == 1) {
            handler = getCommandHandler(start);
            if (handler->name != NULL)
                rc = (handler->fp)(chptr, end);
        }
        exit_status = rc;
        start = end + 1;
    }

    free(contents);
    return rc;
}

void delayOnSignal(int signum) {
    if (ocPid != -1)
        kill(ocPid, SIGSTOP);
    udelay(500000);
    if (ocPid != -1)
        kill(ocPid, SIGCONT);

    signal(signum, delayOnSignal);
}

static void 
sendParentAlarm(nashContext *nc, int usec)
{
    if (nc->hp_parent_pid != -1)
        kill(nc->hp_parent_pid, SIGALRM);
}


static void traceback(int signum, siginfo_t *si, void *data)
{
    void *array[40];
    size_t size;
    struct sigaction act;

    /* XXX PJFIX  we really should not be referencing this if glibc isn't
     * declaring it, but here we are... */
    extern const char *const sys_sigabbrev[NSIG];

    memset(array, '\0', sizeof (array));
    size = backtrace(array, 40);

    fprintf(stderr, "nash received SIG%s!  Backtrace (%ld):\n",
            sys_sigabbrev[signum], (unsigned long)size);
    backtrace_symbols_fd(array, size, STDERR_FILENO);

    memset(&act, '\0', sizeof(act));
    sigemptyset(&act.sa_mask);
    sigaddset(&act.sa_mask, SIGABRT);
    sigaddset(&act.sa_mask, SIGSEGV);
    if (signum == SIGSEGV) {
        act.sa_handler = SIG_DFL;
    } else {
        act.sa_sigaction = traceback;
        act.sa_flags = SA_SIGINFO;
    }
    sigaction(signum, &act, NULL);
    if (signum == SIGSEGV)
        udelay(-1);
}

int main(int argc, char **argv) {
    int fd = 0;
    char * name;
    int rc;
    int force = 0;
    struct sigaction act;
    
    memset(&act, '\0', sizeof(act));
    act.sa_handler = NULL;
    act.sa_sigaction = traceback;
    act.sa_flags = SA_SIGINFO;

    sigemptyset(&act.sa_mask);
    sigaddset(&act.sa_mask, SIGABRT);
    sigaddset(&act.sa_mask, SIGSEGV);
    sigaction(SIGSEGV, &act, NULL);
    sigaction(SIGABRT, &act, NULL);

    _nash_context = nashNewContext();
    nashSetDelayParent(_nash_context, sendParentAlarm);

    name = strrchr(argv[0], '/');
    if (!name)
        name = argv[0];
    else
        name++;

    if (!strcmp(name, "modprobe")) {
        exit(0);
    }

    _nash_context->testing = (getppid() != 0) && (getppid() != 1);
    argv++, argc--;

    while (argc && **argv == '-') {
        if (!strcmp(*argv, "--forcequiet")) {
            force = 1;
            _nash_context->quiet = 1;
            argv++, argc--;
            _nash_context->testing = 0;
        } else if (!strcmp(*argv, "--force")) {
            force = 1;
            argv++, argc--;
            _nash_context->testing = 0;
        } else if (!strcmp(*argv, "--noisy")) {
            _nash_context->quiet = 0;
            _nash_context->reallyquiet = 0;
            argv++, argc--;
        } else if (!strcmp(*argv, "--quiet")) {
            _nash_context->quiet = 1;
            argv++, argc--;
        } else if (!strcmp(*argv, "--reallyquiet")) {
            _nash_context->reallyquiet = 1;
            argv++, argc--;
        } else {
            eprintf("unknown argument %s\n", *argv);
            nashFreeContext(_nash_context);
            return 1;
        }
    }

    if (getpid() == 1) {
        if (force)
            qprintf("(forcing normal run)\n");
        
        if (_nash_context->testing)
            qprintf("(running in test mode).\n");
        
        qprintf("Red Hat nash version %s starting\n", VERSION);
    }

    if (*argv) {
        fd = open(*argv, O_RDONLY, 0);
        if (fd < 0) {
            eprintf("nash: cannot open %s: %m\n", *argv);
            nashFreeContext(_nash_context);
            exit(1);
        }
    }

    signal(SIGALRM, delayOnSignal);

    /* runStartup closes fd */
    rc = runStartup(fd, *argv);

    nashHotplugKill(_nash_context);
    nashFreeContext(_nash_context);
    return rc;
}

/*
 * vim:ts=8:sw=4:sts=4:et
 */
