/*
 * nash-network.c
 *
 * Simple network bring-up code for nash.
 *
 * Copyright 2006, 2007, 2008  Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author(s): Jeremy Katz <katzj@redhat.com>
 *            David Cantrell <dcantrell@redhat.com>
 */

#include <errno.h>
#include <resolv.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <popt.h>

#include <netlink/netlink.h>
#include <netlink/socket.h>
#include <netlink/route/rtnl.h>
#include <netlink/route/route.h>
#include <netlink/route/addr.h>
#include <netlink/route/link.h>

#include <net/if.h>

#include <asm/types.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <sys/socket.h>

#include <errno.h>

#include <sys/ioctl.h>
#include <linux/sockios.h>

#include <nash.h>
#include "util.h"

#define DHCLIENT       "/sbin/dhclient"
#define RESOLVCONF     "/etc/resolv.conf"
#define DHCLIENTLEASES "/var/lib/dhclient/dhclient.leases"
#define DHCLIENTPID    "/var/run/dhclient.pid"
#define BUFSIZE		4096

/* returns 1 for link, 0 for no link, -1 for unknown */
int get_link_status(char *ifname);

#define NUM_LINK_CHECKS 5
static int waitForLink(char * dev) {
    int tries = 0;

    /* try to wait for a valid link -- if the status is unknown or
     * up continue, else sleep for 1 second and try again for up
     * to five times */
    nashLogger(_nash_context, NASH_NOTICE, "waiting for link... ");
    while (tries < NUM_LINK_CHECKS) {
      if (get_link_status(dev) != 0)
            break;
        sleep(1);
        tries++;
    }
    nashLogger(_nash_context, NASH_NOTICE, "%d seconds.\n", tries);
    if (tries < NUM_LINK_CHECKS)
        return 0;
    nashLogger(_nash_context, NASH_WARNING, "no network link detected on %s\n", dev);
    return 1;
}

static void writeResolvConf(char *dns, char *nameserver, char *domain) {
    FILE * f;
    char * tmpdns = NULL, * c = NULL;
    struct in_addr addr;
    struct in6_addr addr6;

    if (dns == NULL && nameserver == NULL && domain == NULL)
        return;

    f = fopen(RESOLVCONF, "w");
    if (!f) {
        eprintf("Cannot create %s: %s\n", RESOLVCONF, strerror(errno));
        return;
    }

    if (domain != NULL)
        fprintf(f, "search %s\n", domain);

    if (nameserver != NULL) {
        if (inet_pton(AF_INET, nameserver, &addr) > 0 ||
            inet_pton(AF_INET6, nameserver, &addr6) > 0) {
            fprintf(f, "nameserver %s\n", nameserver);
        }
    }

    /* dns is a command delimited list of name IP addresses
     * can be IPv4 or IPv6 addresses, just make sure they
     * satisfy input to inet_pton()
     */
    if (dns && *dns) {
        if ((tmpdns = strdup(dns)) == NULL) {
            eprintf("Failure reading dns servers: %s\n", strerror(errno));
            fclose(f);
            return;
        }

        c = strtok(tmpdns, ",");
        while (c != NULL) {
            memset(&addr, 0, sizeof(addr));
            memset(&addr6, 0, sizeof(addr6));

            if (inet_pton(AF_INET, c, &addr) > 0 ||
                inet_pton(AF_INET6, c, &addr) > 0)
                fprintf(f, "nameserver %s\n", c);
            else
                eprintf("Invalid nameserver address %s: %s\n", c, strerror(errno));

            c = strtok(NULL, ",");
        }
    }

    fclose(f);

    res_init();         /* reinit the resolver so DNS changes take affect */

    return;
}

static int nashSetMTU(char * dev, int mtu) {
    struct nl_handle *handle = NULL;
    struct nl_cache *cache = NULL;
    struct rtnl_link *link = NULL;
    struct rtnl_link *request = NULL;
    int ret;

    if (dev == NULL || mtu <= 0)
        return 1;

    if ((handle = nl_handle_alloc()) == NULL)
        return 1;

    if (nl_connect(handle, NETLINK_ROUTE)) {
        nl_handle_destroy(handle);
        return 1;
    }

    if ((cache = rtnl_link_alloc_cache(handle)) == NULL) {
        nl_close(handle);
        nl_handle_destroy(handle);
        return 1;
    }

    if ((link = rtnl_link_get_by_name(cache, dev)) == NULL) {
        nl_cache_free(cache);
        nl_close(handle);
        nl_handle_destroy(handle);
        return 1;
    }

    request = rtnl_link_alloc();
    rtnl_link_set_mtu(request, mtu);

    ret = rtnl_link_change(handle, link, request, 0);

    rtnl_link_put(link);
    rtnl_link_put(request);
    nl_cache_free(cache);
    nl_close(handle);
    nl_handle_destroy(handle);

    return ret;
}

static int nashSetUp(char * dev) {
    struct nl_handle *handle = NULL;
    struct nl_cache *cache = NULL;
    struct rtnl_link *link = NULL;
    struct rtnl_link *request = NULL;
    int ret;

    if (dev == NULL)
        return 1;

    if ((handle = nl_handle_alloc()) == NULL)
        return 1;

    if (nl_connect(handle, NETLINK_ROUTE)) {
        nl_handle_destroy(handle);
        return 1;
    }

    if ((cache = rtnl_link_alloc_cache(handle)) == NULL) {
        nl_close(handle);
        nl_handle_destroy(handle);
        return 1;
    }

    if ((link = rtnl_link_get_by_name(cache, dev)) == NULL) {
        nl_cache_free(cache);
        nl_close(handle);
        nl_handle_destroy(handle);
        return 1;
    }

    request = rtnl_link_alloc();
    rtnl_link_set_flags(request, rtnl_link_str2flags("up"));

    ret = rtnl_link_change(handle, link, request, 0);

    rtnl_link_put(link);
    rtnl_link_put(request);
    nl_cache_free(cache);
    nl_close(handle);
    nl_handle_destroy(handle);

    return ret;
}

static char *dhclientLeaseLineParse(char *linebuf, char *key) {
    char *b = NULL, *e = NULL;

    if ((b = index(strstr(linebuf, key), ' ')) == NULL)
        return NULL;
    b++;

    while (*b == '"')
        b++;

    if ((e = rindex(b, ';')) == NULL)
        return NULL;
    *e = '\0';
    e--;

    while (*e == '"') {
        *e = '\0';
        e--;
    }

    return b;
}

static void nashSetHostname(char *hostname, char *domain) {
    int hfailed = 0;
    char *tmp = NULL;

    if (hostname == NULL)
        return;

    if (domain != NULL) {
        if (asprintf(&tmp, "%s.%s", hostname, domain) != -1) {
            if (sethostname(tmp, strlen(tmp)) == -1) {
                hfailed = 1;
            }

            free(tmp);
        }
    } else {
        if (sethostname(hostname, strlen(hostname)) == -1) {
            return;
        }
    }

    if (hfailed == 1) {
        if (sethostname(hostname, strlen(hostname)) == -1) {
            return;
        }
    }

    return;
}

static char *nashSetupInterface(char *dev, char *ip, char *netmask, char *broadcast, char *gateway) {
    int ifindex = 0;
    int prefix = -1;
    struct nl_handle *handle = NULL;
    struct nl_cache *cache = NULL;
    struct rtnl_addr *addr = NULL;
    struct rtnl_route *route = NULL;
    struct nl_addr *nl_ip = NULL;
    struct nl_addr *nl_bc = NULL;
    struct nl_addr *nl_gw = NULL;
    struct in_addr mask;

    if (dev == NULL || ip == NULL)
        return "dev and ip are empty";

    if ((handle = nl_handle_alloc()) == NULL)
        return "failed to get a Netlink handle";

    if (nl_connect(handle, NETLINK_ROUTE)) {
        nl_handle_destroy(handle);
        return "failed to connect to NETLINK_ROUTE";
    }

    if ((cache = rtnl_link_alloc_cache(handle)) == NULL) {
        nl_close(handle);
        nl_handle_destroy(handle);
        return "failed to get a link cache";
    }

    /* configure the IP address, netmask, broadcast */
    addr = rtnl_addr_alloc();
    ifindex = rtnl_link_name2i(cache, dev);
    rtnl_addr_set_ifindex(addr, ifindex);

    nl_ip = nl_addr_parse(ip, AF_UNSPEC);
    rtnl_addr_set_local(addr, nl_ip);

    if (broadcast != NULL) {
        nl_bc = nl_addr_parse(broadcast, AF_UNSPEC);
        rtnl_addr_set_broadcast(addr, nl_bc);
    }

    if (netmask != NULL) {
        if (inet_pton(AF_INET, netmask, &mask) > 0) {
            unsigned int s_addr = ntohl(mask.s_addr);
            prefix = 32;
            while (!(s_addr & 1)) {
                s_addr >>= 1;
                prefix--;
            }
        }

        if (prefix != -1) {
            rtnl_addr_set_prefixlen(addr, prefix);
        }
    }

    rtnl_addr_add(handle, addr, 0);
    rtnl_addr_put(addr);

    if (nl_ip != NULL) {
        nl_addr_destroy(nl_ip);
    }

    if (nl_bc != NULL) {
        nl_addr_destroy(nl_bc);
    }

    /* configure the gateway */
    if (gateway != NULL) {
        route = rtnl_route_alloc();
        rtnl_route_set_scope(route, RT_SCOPE_UNIVERSE);
        nl_gw = nl_addr_parse(gateway, AF_UNSPEC);
        rtnl_route_set_gateway(route, nl_gw);

        rtnl_route_add(handle, route, 0);
        rtnl_route_put(route);

        if (nl_gw != NULL) {
            nl_addr_destroy(nl_gw);
        }
    }

    /* success */
    nl_cache_free(cache);
    nl_close(handle);
    nl_handle_destroy(handle);
    return NULL;
}

int nashNetworkCommand(int argc, char *argv[]) {
    char * bootProto = NULL, * dev = NULL, * dhcpclass = NULL, * ethtool = NULL, * hostname = NULL;
    char * gateway = NULL, * ip = NULL, * nameserver = NULL, * netmask = NULL, * dns = NULL, * domain = NULL;
    char * broadcast = NULL;
    int mtu = 0, rc, status;
    char * err = NULL;
    char * val = NULL;
    pid_t pid;
    FILE *f = NULL;
    int in_lease = 0;
    int in_interface = 0;
    char linebuf[BUFSIZE];
    poptContext optCon;
    struct poptOption netOptions[] = {
        { "bootproto", '\0', POPT_ARG_STRING, &bootProto, 0, NULL, NULL },
        { "device", '\0', POPT_ARG_STRING, &dev, 0, NULL, NULL },
        { "dhcpclass", '\0', POPT_ARG_STRING, &dhcpclass, 0, NULL, NULL },
        { "dns", '\0', POPT_ARG_STRING, &dns, 0, NULL, NULL },
        { "domain", '\0', POPT_ARG_STRING, &domain, 0, NULL, NULL },
        { "gateway", '\0', POPT_ARG_STRING, &gateway, 'g', NULL, NULL },
        { "ip", '\0', POPT_ARG_STRING, &ip, 'i', NULL, NULL },
        { "nameserver", '\0', POPT_ARG_STRING, &nameserver, 'n', NULL, NULL },
        { "netmask", '\0', POPT_ARG_STRING, &netmask, 'm', NULL, NULL },
        { "ethtool", '\0', POPT_ARG_STRING, &ethtool, 0, NULL, NULL },
        { "mtu", '\0', POPT_ARG_INT, &mtu, 0, NULL, NULL },
        { "hostname", '\0', POPT_ARG_STRING, &hostname, 0, NULL, NULL },
        { NULL, 0, 0, NULL, 0, NULL, NULL }
    };

    optCon = poptGetContext(NULL, argc, (const char **) argv,
                            netOptions, 0);

    while ((rc = poptGetNextOpt(optCon)) > 0) {}
    if (rc < -1) {
        nashLogger(_nash_context, NASH_ERROR, "ERROR: Bad argument to network command\n");
        return 1;
    }

    if (ethtool != NULL) {
        /* FIXME */
        nashLogger(_nash_context, NASH_WARNING, "WARNING: ethtool options not currently handled\n");
    }

    if (dev == NULL)
        dev = strdup("eth0");

    if (mtu) {
        if (nashSetMTU(dev, mtu)) {
            nashLogger(_nash_context, NASH_WARNING, "WARNING: unable to set MTU on %s to %d\n", dev, mtu);
        }
    }

    if (hostname)
        nashSetHostname(hostname, domain);

    if (dns || nameserver)
        writeResolvConf(dns, nameserver, domain);

    if (nashSetUp(dev)) {
        eprintf("ERROR: Could not bring interface %s up\n", dev);
        return 1;
    }

    if ((bootProto != NULL) && (!strncmp(bootProto, "dhcp", 4))) {
        nashLogger(_nash_context, NASH_NOTICE, "Sending request for IP information through %s\n", dev);

        waitForLink(dev);
        pid = fork();

        /* start dhclient */
        if (pid == 0) {
            /* dhclient returns code 2 for failure, use a different code to
               signal we never actually managed to start dhclient */
            if (setpgrp() == -1)
                exit(3);

            if (execl(DHCLIENT, DHCLIENT, "-1", "-sf", "/dev/null", 
                    "-R", "subnet-mask,broadcast-address,time-offset,routers,domain-name,domain-name-servers,host-name,nis-domain,nis-servers,ntp-servers,root-path",
                    dev, NULL) == -1)
                exit(3);
        } else if (pid == -1) {
            eprintf("Failed to fork to start dhclient: %m\n");
            return 1;
        } else {
            if (waitpid(pid, &status, 0) == -1) {
                eprintf("waitpid() for dhclient failed: %m\n");
                return 1;
            }
            if (!WIFEXITED(status)) {
                eprintf("dhclient did not exit normally\n");
                return 1;
            }
            if (WEXITSTATUS(status) != 0) {
                if (WEXITSTATUS(status) == 2)
                    eprintf("Could not get a dhcp lease\n");
                else
                    eprintf("dhclient abnormal exit status: %d\n",
                            WEXITSTATUS(status));
                return 1;
            }
        }

        /* collect lease information */
        if ((f = fopen(DHCLIENTLEASES, "r")) == NULL) {
            eprintf("Failed to open %s: %m\n", DHCLIENTLEASES);
            return 1;
        }

        while (fgets(linebuf, BUFSIZE, f) != NULL) {
            linebuf[strlen(linebuf) - 1] = '\0';

            if (in_lease == 0) {
                if (strstr(linebuf, "lease {") != NULL) {
                    in_lease = 1;
                }
            } else if (in_lease == 1) {
                if (strstr(linebuf, "interface") != NULL) {
                    val = dhclientLeaseLineParse(linebuf, "interface");

                    if (!strcmp(dev, val)) {
                        in_lease = 2;
                        in_interface = 1;
                    }
                }
            } else if (in_interface == 1) {
                if (strstr(linebuf, "fixed-address") != NULL) {
                    val = dhclientLeaseLineParse(linebuf, "fixed-address");
                    ip = strdup(val);
                } else if (strstr(linebuf, "subnet-mask") != NULL) {
                    val = dhclientLeaseLineParse(linebuf, "subnet-mask");
                    netmask = strdup(val);
                } else if (strstr(linebuf, "domain-name-servers") != NULL) {
                    val = dhclientLeaseLineParse(linebuf, "domain-name-servers");
                    dns = strdup(val);
                    val = dns;

                    while (*val != '\0') {
                        if (*val == ' ')
                            *val = ',';
                        val++;
                    }
                } else if (strstr(linebuf, "broadcast-address") != NULL) {
                    val = dhclientLeaseLineParse(linebuf, "broadcast-address");
                    broadcast = strdup(val);
                } else if (strstr(linebuf, "routers") != NULL) {
                    val = dhclientLeaseLineParse(linebuf, "routers");
                    gateway = strdup(val);
                } else if (strstr(linebuf, "host-name") != NULL) {
                    val = dhclientLeaseLineParse(linebuf, "host-name");
                    hostname = strdup(val);
                } else if (strstr(linebuf, "domain-name") != NULL) {
                    val = dhclientLeaseLineParse(linebuf, "domain-name");
                    domain = strdup(val);
                }

                if (strstr(linebuf, "}") != NULL) {
                    in_lease = 0;
                    in_interface = 0;
                }
            }
        }

        fclose(f);

        /* Kill the background dhclient process */
        if ((f = fopen(DHCLIENTPID, "r"))) {
            if(fgets(linebuf, BUFSIZE, f)) {
                pid = strtol(linebuf, NULL, 10);
                kill(pid, SIGTERM);
            }
            fclose(f);
        }

        if (hostname)
            nashSetHostname(hostname, domain);
    }

    err = nashSetupInterface(dev, ip, netmask, broadcast, gateway);
    if (err) {
        nashLogger(_nash_context, NASH_ERROR, "ERROR: Interface setup failed: %s\n", err);
        return 1;
    }
    writeResolvConf(dns, nameserver, domain);

    sleep(2);
    return 0;
}

struct nl {
    int fd;
    struct sockaddr_nl addr;
    int seq;
};

static int
rtnl_wilddump_request(struct nl *nl, int family, int type)
{
    struct sockaddr_nl nladdr;
    struct {
        struct nlmsghdr nh;
        struct rtgenmsg g;
    } req;
    int status;

    memset(&nladdr, '\0', sizeof(nladdr));
    nladdr.nl_family = AF_NETLINK;

    memset(&req, '\0', sizeof(req));
    req.nh.nlmsg_len = sizeof(req);
    req.nh.nlmsg_flags = NLM_F_DUMP | NLM_F_REQUEST ;
    req.nh.nlmsg_type = type;
    req.nh.nlmsg_seq = ++nl->seq;
    req.nh.nlmsg_pid = 0;
    req.g.rtgen_family = family;

    status = sendto(nl->fd, &req, sizeof(req), 0, (struct sockaddr *)&nladdr,
            sizeof(nladdr));
    if (status < 0) {
        eprintf("rtnl_wilddump_request: sendto failed: %m\n");
        return -1;
    }

    return 0;
}

struct ifinfo {
    int index;
    char *name;
    char *mac;
    struct ifinfo *next;
};

static int
rtnl_dump_recv(struct nl *nl, struct ifinfo **iftable)
{
    struct sockaddr_nl nladdr;
    struct iovec iov;
    struct msghdr msg = {
        .msg_name = &nladdr,
        .msg_namelen = sizeof(nladdr),
        .msg_iov = &iov,
        .msg_iovlen = 1,
    };
    char buf[16384];
    struct ifinfo *ifentry = NULL;

    iov.iov_base = buf;
    while (1) {
        int status;
        struct nlmsghdr *h;

        memset(buf, '\0', sizeof(buf));
        iov.iov_len = sizeof(buf);
        status = recvmsg(nl->fd, &msg, 0);

        if (status < 0) {
            if (errno == EINTR)
                continue;
            eprintf("rtnl_dump_recv: recvmsg failed: %m\n");
            return -1;
        }

        if (status == 0) {
            eprintf("rtnl_dump_recv: received EOF\n");
            return -1;
        }

        h = (struct nlmsghdr *)buf;
        while (NLMSG_OK(h, status)) {
            struct rtattr *rta;
            struct ifinfomsg *ifi;
            int len;

            if (h->nlmsg_type == NLMSG_DONE)
                return 0;

            if (nladdr.nl_pid != 0 || h->nlmsg_pid != nl->addr.nl_pid ||
                    h->nlmsg_seq != nl->seq)
                goto skip_it;

            if (h->nlmsg_type == NLMSG_ERROR)
                goto skip_it;

            if (h->nlmsg_type == RTM_NEWLINK) {
                ifentry = calloc(1, sizeof(*ifentry));

                ifi = NLMSG_DATA(h);
                ifentry->index = ifi->ifi_index;
                len = IFLA_PAYLOAD(h);
                rta = IFLA_RTA(ifi);
                while (RTA_OK(rta, len)) {
                    switch(rta->rta_type) {
                        case IFLA_IFNAME:
                            ifentry->name = strdup(RTA_DATA(rta));
                            break;
                        case IFLA_ADDRESS: {
                            int alen = RTA_PAYLOAD(rta);
                            int i;
                            char *mac = calloc(alen * 2 + alen, sizeof(char));
                            char *data, *octet;

                            data = RTA_DATA(rta);
                            octet = mac;

                            for (i = 0; i < alen; i++) {
                                sprintf(octet, "%02x:", (unsigned char)data[i]);
                                octet += 3;
                            }
                            mac[i*2+i-1] = '\0';
                            ifentry->mac = mac;
                            break;
                        }
                        default:
                            break;
                    }
                    rta = RTA_NEXT(rta,len);
                }
                ifentry->next = *iftable;
                *iftable = ifentry;
            }
            
skip_it:
            h = NLMSG_NEXT(h, status);
        }
    }

    return 0;
}

static int
get_ifinfo(struct ifinfo **table)
{
    struct nl nl;
    socklen_t socklen;
    int n;
    int rc;

    memset(&nl, '\0', sizeof(nl));
    nl.addr.nl_family = AF_NETLINK;
    nl.seq = time(NULL);

    nl.fd = socket(PF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
    if (nl.fd < 0) {
        eprintf("socket() failed: %m\n");
        return -1;
    }

    n = 16384;
    socklen = sizeof(n);
    setsockopt(nl.fd, SOL_SOCKET, SO_SNDBUF, &n, socklen);
    n = 16384;
    setsockopt(nl.fd, SOL_SOCKET, SO_RCVBUF, &n, socklen);
    bind(nl.fd, (struct sockaddr *)&nl.addr, sizeof(nl.addr));

    memset(&nl.addr, 0, sizeof(nl.addr));
    socklen = sizeof(nl.addr);
    getsockname(nl.fd, (struct sockaddr *)&nl.addr, &socklen);

    rc = rtnl_wilddump_request(&nl, AF_UNSPEC, RTM_GETLINK);
    if (rc >= 0)
        rc = rtnl_dump_recv(&nl, table);

    close(nl.fd);
    return rc;
}

static void
free_ifinfo(struct ifinfo *table)
{
    struct ifinfo *ifentry = table;

    while (ifentry) {
        struct ifinfo *next = ifentry->next;

        if (ifentry->name)
            free(ifentry->name);
        if (ifentry->mac)
            free(ifentry->mac);
        free(ifentry);
        ifentry = next;
    }
}

static int
get_ctl_fd(void)
{
	int s_errno;
	int fd;

	fd = socket(PF_INET, SOCK_DGRAM, 0);
	if (fd >= 0)
		return fd;
	s_errno = errno;
	fd = socket(PF_PACKET, SOCK_DGRAM, 0);
	if (fd >= 0)
		return fd;
	fd = socket(PF_INET6, SOCK_DGRAM, 0);
	if (fd >= 0)
		return fd;
	errno = s_errno;
	eprintf("Cannot create control socket: %m\n");
	return -1;
}

static int
do_changename(const char *dev, const char *newdev)
{
	struct ifreq ifr;
	int fd;
	int err, en;

	strncpy(ifr.ifr_name, dev, IFNAMSIZ);
	strncpy(ifr.ifr_newname, newdev, IFNAMSIZ);
	fd = get_ctl_fd();
	if (fd < 0)
		return -1;
	err = ioctl(fd, SIOCSIFNAME, &ifr);
        en = errno;
	close(fd);
        errno = en;
        if (err)
        	eprintf("Change Name failed: %m\n");
        return err;
}

int
renameNetworkInterface(char *hwaddr, char *ifname)
{
    struct ifinfo *ifinfo = NULL, *ifentry;
    int rc = 0;

    int collision = 0;
    char *oldname = NULL;
    
    /* if we're called like "netname $BOOTIF eth0", that'll expand to
     * "01-c0-ff-ee-c0-ff-ee" .  We don't want the leading "01-", and
     * we want colons not dashes. */
    if (strlen(hwaddr) == 20 && !strncmp(hwaddr, "01-", 3)) {
        char *macsep;
        hwaddr += 3;

        macsep = hwaddr;
        while ((macsep = strchr(macsep, '-')))
            *macsep = ':';
    }

    if (get_ifinfo(&ifinfo) >= 0) {
        for (ifentry = ifinfo; ifentry; ifentry = ifentry->next) {
            if (ifentry->mac && !strcasecmp(hwaddr, ifentry->mac)) {
                if (!strcmp(ifname, ifentry->name))
                    goto leave; /* no renaming needed */
                oldname = ifentry->name;
            } else if (!strcmp(ifname, ifentry->name)) {
                collision = 1;
            }
        }

        if (!oldname) {
            eprintf("No interface with MAC: %s found\n", hwaddr);
            rc = -1;
            goto leave;
        }

        if (collision)
            rc = do_changename(ifname, "nashdevname");
        if (rc >= 0)
            rc = do_changename(oldname, ifname);
        if (rc >= 0 && collision)
            rc = do_changename("nashdevname", oldname);

leave:
        free_ifinfo(ifinfo);
    }

    return rc;
}

